# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import logging
from typing import Dict, List, Optional

from simplejson.scanner import JSONDecodeError

from snapcraft import formatting_utils
from snapcraft.internal.errors import SnapcraftError, SnapcraftException

from . import channels, status

logger = logging.getLogger(__name__)

_FORUM_URL = "https://forum.snapcraft.io/c/store"


class StoreError(SnapcraftError):
    """Base class for all storeapi exceptions.

    :cvar fmt: A format string that daughter classes override

    """

    def __init__(self, **kwargs):
        with contextlib.suppress(KeyError, AttributeError):
            logger.debug("Store error response: {}".format(kwargs["response"].__dict__))
        super().__init__(**kwargs)


class GeneralStoreError(StoreError):

    fmt = "Store Error: {message}"

    def __init__(self, message, response=None):
        # Unfortunately the store doesn't give us a consistent error response,
        # so we'll check the ones of which we're aware.
        with contextlib.suppress(AttributeError, JSONDecodeError):
            response_json = response.json()
            extra_error_message = ""
            if "error_message" in response_json:
                extra_error_message = response_json["error_message"]
            elif "message" in response_json:
                extra_error_message = response_json["message"]

            if extra_error_message:
                message += ": {}".format(extra_error_message)

        super().__init__(response=response, message=message)


class StoreErrorList:
    def __str__(self) -> str:
        error_list: List[str] = list()
        for error in self._error_list:
            error_list.append("- {}: {}".format(error["code"], error["message"]))
        return "\n".join(error_list).strip()

    def __repr__(self) -> str:
        code_list = []
        for error in self._error_list:
            code = error.get("code")
            if code:
                code_list.append(code)

        return "<StoreErrorList: {}>".format(" ".join(code_list))

    def __contains__(self, error_code: str) -> bool:
        return any((error.get("code") == error_code for error in self._error_list))

    def __getitem__(self, error_code: str) -> Dict[str, str]:
        for error in self._error_list:
            if error.get("code") == error_code:
                return error

        raise KeyError(error_code)

    def __init__(self, error_list: List[Dict[str, str]]) -> None:
        self._error_list = error_list


class SnapNotFoundError(SnapcraftException):
    def __init__(
        self,
        *,
        snap_name: str = "",
        snap_id: Optional[str] = None,
        channel: Optional[str] = None,
        arch: Optional[str] = None,
    ):
        super().__init__()

        # Defaulting snap_name to "" to support the one case we have that
        # makes use of snap_id.
        if snap_name == "" and snap_id is None:
            raise RuntimeError("Both 'snap_name' and 'snap_id' cannot be None.")

        self._snap_name = snap_name
        self._snap_id = snap_id
        self._channel = channel
        self._arch = arch

    def get_brief(self) -> str:
        if self._snap_id:
            # This is legacy.
            brief = f"Cannot find snap with snap_id {self._snap_id!r}."
        elif self._channel and self._arch:
            brief = f"Snap {self._snap_name!r} for architecture {self._arch!r} was not found on channel {self._channel!r}."
        elif self._channel:
            brief = (
                f"Snap {self._snap_name!r} was not found on channel {self._channel!r}."
            )
        elif self._arch:
            brief = f"Snap {self._snap_name!r} for architecture {self._arch!r} was not found."
        else:
            brief = f"Snap {self._snap_name!r} was not found."

        return brief

    def get_resolution(self) -> str:
        # All new APIs use snap_name.
        if self._snap_id:
            snap_identifier = self._snap_id
        else:
            snap_identifier = self._snap_name

        resolution = f"Ensure you have proper access rights for {snap_identifier!r}."

        if self._channel and self._arch:
            resolution += (
                f"\nAlso ensure the correct channel and architecture was used."
            )
        elif self._channel:
            resolution += f"\nAlso ensure the correct channel was used."
        elif self._arch:
            resolution += f"\nAlso ensure the correct architecture was used."

        return resolution


class NoSnapIdError(StoreError):

    fmt = (
        "Failed to get snap ID for snap {snap_name!r}. This is an error in "
        "the store, please open a new topic in the 'store' category in the "
        "forum: {forum_url}"
    )

    def __init__(self, snap_name):
        super().__init__(snap_name=snap_name, forum_url=_FORUM_URL)


class StoreDownloadError(StoreError):
    pass


class DownloadNotFoundError(StoreDownloadError):

    fmt = "Downloaded file not found {path!r}."

    def __init__(self, *, path: str) -> None:
        super().__init__(path=path)


class SHAMismatchError(StoreDownloadError):

    fmt = (
        "The SHA3-384 checksum for {path!r} was {calculated!r}: expected {expected!r}."
    )

    def __init__(self, *, path: str, expected: str, calculated: str) -> None:
        super().__init__(path=path, expected=expected, calculated=calculated)


class DeveloperAgreementSignError(StoreError):

    fmt = (
        "There was an error while signing developer agreement.\n"
        "Reason: {reason!r}\n"
        "Text: {text!r}"
    )

    def __init__(self, response):
        super().__init__(response=response, reason=response.reason, text=response.text)


class NeedTermsSignedError(StoreError):

    fmt = (
        "Developer Terms of Service agreement must be signed "
        "before continuing: {message}"
    )

    def __init__(self, message):
        super().__init__(message=message)


class StoreAccountInformationError(StoreError):

    fmt = "Error fetching account information from store: {error}"

    def __init__(self, response):
        error = "{} {}".format(response.status_code, response.reason)
        extra = []
        try:
            response_json = response.json()
            if "error_list" in response_json:
                error = " ".join(
                    error["message"] for error in response_json["error_list"]
                )
                extra = [
                    error["extra"]
                    for error in response_json["error_list"]
                    if "extra" in error
                ]
        except JSONDecodeError:
            pass
        super().__init__(response=response, error=error, extra=extra)


class StoreKeyRegistrationError(StoreError):

    fmt = "Key registration failed: {error}"

    def __init__(self, response):
        error = "{} {}".format(response.status_code, response.reason)
        try:
            response_json = response.json()
            if "error_list" in response_json:
                error = " ".join(
                    error["message"] for error in response_json["error_list"]
                )
        except JSONDecodeError:
            pass
        super().__init__(response=response, error=error)


class StoreRegistrationError(StoreError):
    """Captures store name registration errors.

    Overrides `SnapcraftError` setup and string representation to
    support new-style (multiple) Store error responses.

    See https://myapps.developer.ubuntu.com/docs/api/snap.html#errors
    """

    __FMT_ALREADY_REGISTERED = (
        "The name {snap_name!r} is already taken.\n\n"
        "We can if needed rename snaps to ensure they match the expectations "
        "of most users. If you are the publisher most users expect for "
        "{snap_name!r} then claim the name at {register_name_url!r}"
    )

    __FMT_ALREADY_OWNED = "You already own the name {snap_name!r}."

    __FMT_RESERVED = (
        "The name {snap_name!r} is reserved.\n\n"
        "If you are the publisher most users expect for "
        "{snap_name!r} then please claim the name at {register_name_url!r}\n\n"
        "Otherwise, please register another name."
    )

    __FMT_RETRY_WAIT = (
        "You must wait {retry_after} seconds before trying to register "
        "your next snap."
    )

    __FMT_INVALID = "{message}"

    fmt = "Registration failed."

    __error_messages = {
        "already_registered": __FMT_ALREADY_REGISTERED,
        "already_owned": __FMT_ALREADY_OWNED,
        "reserved_name": __FMT_RESERVED,
        "register_window": __FMT_RETRY_WAIT,
        "invalid": __FMT_INVALID,
    }

    def __init__(self, snap_name, response):
        self._errors = []

        try:
            response_json = response.json()
        except JSONDecodeError:
            response_json = {}

        # Annotate 'snap_name' to be used when formatting errors.
        response_json["snap_name"] = snap_name

        # Extract the new-format error structure.
        error_list = response_json.pop("error_list", None)

        # Cope with legacy errors (missing 'error_list', single error and
        # top-level 'code').
        if error_list is None:
            error_code = response_json.get("code")
            # we default to self.fmt in case error_code is not mapped yet.
            fmt = self.__error_messages.get(error_code, self.fmt)
            self._errors.append(fmt.format(**response_json))
            return

        # Support new style formatted errors.
        for error in error_list:
            fmt = self.__error_messages.get(error["code"], self.fmt)
            # Augment 'error' with remaining top-level keys. Should be a
            # no-op once they are moved to their specific error record.
            error.update(response_json)
            self._errors.append(fmt.format(**error))

    def __str__(self):
        """Simply join formatted error as lines."""
        return "\n".join(self._errors)


class StoreUpDownError(StoreError):

    fmt = (
        "There was an error uploading the package.\n"
        "Reason: {reason!r}\n"
        "Text: {text!r}"
    )

    def __init__(self, response):
        super().__init__(response=response, reason=response.reason, text=response.text)


class StoreUploadError(StoreError):

    __FMT_NOT_REGISTERED = (
        "This snap is not registered. Register the snap and try again."
    )

    __FMT_NOT_OWNER = (
        "You are not the publisher or allowed to upload revisions for this "
        "snap. Ensure you are logged in with the proper account and try "
        "again."
    )

    def __init__(self, snap_name, response):
        try:
            response_json = response.json()
        except (AttributeError, JSONDecodeError):
            response_json = {"error_list": list()}

        self.error_list = StoreErrorList(error_list=response_json.pop("error_list"))

        if "resource-forbidden" in self.error_list:
            self.fmt = self.__FMT_NOT_OWNER
        elif "resource-not-found" in self.error_list:
            self.fmt = self.__FMT_NOT_REGISTERED
        else:
            self.fmt = "Received:\n{!s}".format(self.error_list)

        super().__init__(
            response=response,
            snap_name=snap_name,
            status_code=response.status_code,
            **response_json,
        )


class StoreReviewError(StoreError):

    __FMT_NEED_MANUAL_REVIEW = (
        "The Store automatic review failed.\n"
        "A human will soon review your snap, but if you can't wait please "
        "write in the snapcraft forum asking for the manual review "
        "explicitly.\n"
        "If you need to disable confinement, please consider using devmode, "
        "but note that devmode revision will only be allowed to be released "
        "in edge and beta channels.\n"
        "Please check the errors and some hints below:"
    )

    __FMT_PROCESSING_ERROR = "The store was unable to accept this snap."

    __FMT_PROCESSING_UPLOAD_DELTA_ERROR = (
        "There has been a problem while processing a snap delta."
    )

    __messages = {
        "need_manual_review": __FMT_NEED_MANUAL_REVIEW,
        "processing_error": __FMT_PROCESSING_ERROR,
        "processing_upload_delta_error": __FMT_PROCESSING_UPLOAD_DELTA_ERROR,
    }

    def __init__(self, result):
        self.fmt = self.__messages[result["code"]]
        additional = []
        errors = result.get("errors")
        if errors:
            for error in errors:
                message = error.get("message")
                if message:
                    additional.append("  - {message}".format(message=message))
        if additional:
            self.additional = "\n".join(additional)
            self.fmt += "\n{additional}"
        self.code = result["code"]
        super().__init__()


class StoreReleaseError(StoreError):

    fmt = "{message}"

    __FMT_NOT_REGISTERED = (
        "Sorry, try `snapcraft register {snap_name}` before trying to "
        "release or choose an existing revision."
    )

    __FMT_BAD_REQUEST = "{code}: {message}\n"

    __FMT_UNAUTHORIZED_OR_FORBIDDEN = "Received {status_code!r}: {text}"

    def __init__(self, snap_name, response):
        self.fmt_errors = {
            400: self.__fmt_error_400,
            401: self.__fmt_error_401_or_403,
            403: self.__fmt_error_401_or_403,
            404: self.__fmt_error_404,
        }

        fmt_error = self.fmt_errors.get(response.status_code, self.__fmt_error_unknown)

        super().__init__(
            response=response, message=fmt_error(response).format(snap_name=snap_name)
        )

    def __to_json(self, response):
        try:
            response_json = response.json()
        except (AttributeError, JSONDecodeError):
            response_json = {}

        return response_json

    def __fmt_error_400(self, response):
        response_json = self.__to_json(response)
        try:
            fmt = ""
            for error in response_json["error_list"]:
                # Escape curly braces to avoid formatting errors.
                message = error["message"].replace("{", "{{").replace("}", "}}")
                code = error["code"].replace("{", "{{").replace("}", "}}")
                fmt += self.__FMT_BAD_REQUEST.format(code=code, message=message)
            # Strip the last new line from the error message
            fmt = fmt.rstrip("\n")

        except (AttributeError, KeyError):
            fmt = self.__fmt_error_unknown(response)
        return fmt

    def __fmt_error_401_or_403(self, response):
        try:
            text = response.text

            with contextlib.suppress(AttributeError, JSONDecodeError):
                response_json = response.json()
                if "error_list" in response_json:
                    text = _error_list_to_message(response_json)

        except AttributeError:
            text = "error while releasing"

        return self.__FMT_UNAUTHORIZED_OR_FORBIDDEN.format(
            status_code=response.status_code, text=text
        )

    def __fmt_error_404(self, response):
        return self.__FMT_NOT_REGISTERED

    def __fmt_error_unknown(self, response):
        response_json = self.__to_json(response)

        try:
            fmt = "{errors}".format(**response_json)

        except AttributeError:
            fmt = "{}".format(response)

        return fmt


class StoreMetadataError(StoreError):

    __FMT_NOT_FOUND = (
        "Sorry, updating the information on the store has failed, first run "
        "`snapcraft register {snap_name}` and then "
        "`snapcraft upload <snap-file>`."
    )

    fmt = "Received {status_code!r}: {text!r}"

    def __init__(self, snap_name, response, metadata):
        try:
            response_json = response.json()
        except (AttributeError, JSONDecodeError):
            response_json = {}

        if response.status_code == 404:
            self.fmt = self.__FMT_NOT_FOUND
        elif response.status_code == 409:
            conflicts = [
                (error["extra"]["name"], error)
                for error in response_json["error_list"]
                if error["code"] == "conflict"
            ]
            parts = ["Metadata not uploaded!"]
            for field_name, error in sorted(conflicts):
                sent = metadata.get(field_name)
                parts.extend(
                    (
                        "Conflict in {!r} field:".format(field_name),
                        "    In snapcraft.yaml: {!r}".format(sent),
                        "    In the Store:      {!r}".format(error["extra"]["current"]),
                    )
                )
            parts.append(
                "You can repeat the upload-metadata command with "
                "--force to force the local values into the Store"
            )
            self.parts = "\n".join(parts)
            self.fmt = "{parts}"
        elif "error_list" in response_json:
            response_json["text"] = response_json["error_list"][0]["message"]

        super().__init__(
            response=response,
            snap_name=snap_name,
            status_code=response.status_code,
            **response_json,
        )


class StoreValidationError(StoreError):

    fmt = "Received error {status_code!r}: {text!r}"

    def __init__(self, snap_id, response, message=None):
        try:
            response_json = response.json()
            error = response.json()["error_list"][0]
            response_json["text"] = error.get("message")
            response_json["extra"] = error.get("extra")
        except (AttributeError, JSONDecodeError):
            response_json = {"text": message or response}

        super().__init__(
            response=response, status_code=response.status_code, **response_json
        )


class StoreValidationSetsError(StoreError):

    fmt = "Issues encountered with validation set: {error}"

    def __init__(self, response):
        error = "{} {}".format(response.status_code, response.reason)
        try:
            response_json = response.json()

            if "error-list" in response_json:
                error = " ".join(
                    error["message"] for error in response_json["error-list"]
                )
        except JSONDecodeError:
            pass

        super().__init__(response=response, error=error)


class StoreSnapBuildError(StoreError):

    fmt = "Could not assert build: {error}"

    def __init__(self, response):
        error = "{} {}".format(response.status_code, response.reason)
        try:
            response_json = response.json()
            if "error_list" in response_json:
                error = " ".join(
                    error["message"] for error in response_json["error_list"]
                )
        except JSONDecodeError:
            pass

        super().__init__(response=response, error=error)


class StoreSnapRevisionsError(StoreError):

    fmt = (
        "Error fetching revisions of snap id {snap_id!r} for {arch!r} "
        "in {series!r} series: {error}."
    )

    def __init__(self, response, snap_id, series, arch):
        error = "{} {}".format(response.status_code, response.reason)
        try:
            response_json = response.json()
            if "error_list" in response_json:
                error = " ".join(
                    error["message"] for error in response_json["error_list"]
                )
        except JSONDecodeError:
            pass

        super().__init__(
            response=response,
            snap_id=snap_id,
            arch=arch or "any arch",
            series=series or "any",
            error=error,
        )


class StoreDeltaApplicationError(StoreError):

    fmt = "{message}"

    def __init__(self, message):
        super().__init__(message=message)


class StoreSnapChannelMapError(SnapcraftException):
    def __init__(self, *, snap_name: str) -> None:
        self._snap_name = snap_name

    def get_brief(self) -> str:
        return f"Could not retrieve information for {self._snap_name!r}."

    def get_resolution(self) -> str:
        return (
            "Ensure the snap name is correct and that you have permissions to "
            "access it."
        )


class StoreSnapStatusError(StoreSnapRevisionsError):

    fmt = (
        "Error fetching status of snap id {snap_id!r} for {arch!r} "
        "in {series!r} series: {error}."
    )


class StoreChannelClosingError(StoreError):

    fmt = "Could not close channel: {error}"

    def __init__(self, response):
        try:
            e = response.json()["error_list"][0]
            error = "{}".format(e["message"])
        except (JSONDecodeError, KeyError, IndexError):
            error = "{} {}".format(response.status_code, response.reason)

        super().__init__(response=response, error=error)


class StoreChannelClosingPermissionError(StoreError):

    fmt = (
        "Your account lacks permission to close channels for this snap. Make "
        "sure the logged in account has upload permissions on {snap_name!r} "
        "in series {snap_series!r}."
    )

    def __init__(self, snap_name, snap_series):
        super().__init__(snap_name=snap_name, snap_series=snap_series)


class StoreBuildAssertionPermissionError(StoreError):

    fmt = (
        "Your account lacks permission to assert builds for this snap. Make "
        "sure you are logged in as the publisher of {snap_name!r} "
        "for series {snap_series!r}."
    )

    def __init__(self, snap_name, snap_series):
        super().__init__(snap_name=snap_name, snap_series=snap_series)


class StoreAssertionError(StoreError):

    fmt = "Error signing {endpoint} assertion for {snap_name}: {error!s}"


class KeyAlreadyRegisteredError(StoreError):

    fmt = "You have already registered a key named {key_name!r}"

    def __init__(self, key_name):
        super().__init__(key_name=key_name)


class NoKeysError(StoreError):

    fmt = (
        "You have no usable keys.\nPlease create at least one key with "
        "`snapcraft create-key` for use with snap."
    )


class NoSuchKeyError(StoreError):

    fmt = (
        "You have no usable key named {key_name!r}.\nSee the keys available "
        "in your system with `snapcraft keys`."
    )

    def __init__(self, key_name):
        super().__init__(key_name=key_name)


class KeyNotRegisteredError(StoreError):

    fmt = (
        "The key {key_name!r} is not registered in the Store.\nPlease "
        "register it with `snapcraft register-key {key_name!r}` before "
        "signing and uploading signatures to the Store."
    )

    def __init__(self, key_name):
        super().__init__(key_name=key_name)


class InvalidValidationRequestsError(StoreError):

    fmt = "Invalid validation requests (format must be name=revision): {requests}"

    def __init__(self, requests):
        requests_str = " ".join(requests)
        super().__init__(requests=requests_str)


class SignBuildAssertionError(StoreError):

    fmt = "Failed to sign build assertion for {snap_name!r}"

    def __init__(self, snap_name):
        super().__init__(snap_name=snap_name)


def _error_list_to_message(response_json):
    """Handle error list.

    The error format is given here:
    https://dashboard.snapcraft.io/docs/api/snap.html#format
    """
    messages = []
    for error_list_item in response_json["error_list"]:
        messages.append(_error_list_item_to_message(error_list_item, response_json))

    return ", and ".join(messages)


def _error_list_item_to_message(error_list_item, response_json):
    """Handle error list item according to code.

    The list of codes are here:
    https://dashboard.snapcraft.io/docs/api/snap.html#codes
    """
    code = error_list_item["code"]
    message = ""
    if code == "macaroon-permission-required":
        message = _handle_macaroon_permission_required(response_json)

    if message:
        return message
    else:
        return error_list_item["message"]


def _handle_macaroon_permission_required(response_json):
    if "permission" in response_json and "channels" in response_json:
        if response_json["permission"] == "channel":
            channels = response_json["channels"]
            return "Lacking permission to release to channel(s) {}".format(
                formatting_utils.humanize_list(channels, "and")
            )

    return ""


class ChannelNotAvailableOnArchError(StoreError):
    fmt = (
        "No releases available for {snap_name!r} on channel {channel!r} "
        "for architecture {arch!r}.\n"
        "Ensure the selected channel contains released revisions for this architecture."
    )

    def __init__(self, *, snap_name: str, channel: channels.Channel, arch: str) -> None:
        super().__init__(snap_name=snap_name, channel=channel, arch=arch)


class InvalidChannelSet(StoreError):
    fmt = (
        "The {channel!r} channel for {snap_name!r} does not form a complete set.\n"
        "There is no revision released for the following architectures: {arches!r}.\n"
        "Ensure the selected channel contains released revisions for "
        "all architectures."
    )

    def __init__(
        self,
        *,
        snap_name: str,
        channel: channels.Channel,
        channel_outliers: List[status.SnapStatusChannelDetails],
    ) -> None:
        arches = formatting_utils.humanize_list(
            [c.arch for c in channel_outliers], "and"
        )
        super().__init__(snap_name=snap_name, channel=channel, arches=arches)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tarfile
from unittest import mock

import requests
from testtools.matchers import Equals

from snapcraft.internal import sources
from tests import unit


class TestTar(unit.FakeFileHTTPServerBasedTestCase):
    @mock.patch("snapcraft.sources.Tar.provision")
    def test_pull_tarball_must_download_to_sourcedir(self, mock_prov):
        plugin_name = "test_plugin"
        dest_dir = os.path.join("parts", plugin_name, "src")
        os.makedirs(dest_dir)
        tar_file_name = "test.tar"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=tar_file_name
        )
        tar_source = sources.Tar(source, dest_dir)

        tar_source.pull()

        source_file = os.path.join(dest_dir, tar_file_name)
        mock_prov.assert_called_once_with(dest_dir, src=source_file, clean_target=False)
        with open(os.path.join(dest_dir, tar_file_name), "r") as tar_file:
            self.assertThat(tar_file.read(), Equals("Test fake file"))

    @mock.patch("snapcraft.sources.Tar.provision")
    def test_pull_twice_downloads_once(self, mock_prov):
        """If a source checksum is defined, the cache should be tried first."""
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name="test.tar"
        )
        expected_checksum = (
            "sha384/d9da1f5d54432edc8963cd817ceced83f7c6d61d3"
            "50ad76d1c2f50c4935d11d50211945ca0ecb980c04c98099"
            "085b0c3"
        )
        tar_source = sources.Tar(source, self.path, source_checksum=expected_checksum)

        tar_source.pull()
        with mock.patch(
            "requests.get", new=mock.Mock(wraps=requests.get)
        ) as download_spy:
            tar_source.pull()
            self.assertThat(download_spy.call_count, Equals(0))

    def test_strip_common_prefix(self):
        # Create tar file for testing
        os.makedirs(os.path.join("src", "test_prefix"))
        file_to_tar = os.path.join("src", "test_prefix", "test.txt")
        open(file_to_tar, "w").close()
        tar = tarfile.open(os.path.join("src", "test.tar"), "w")
        tar.add(file_to_tar)
        tar.close()

        tar_source = sources.Tar(os.path.join("src", "test.tar"), "dst")
        os.mkdir("dst")
        tar_source.pull()

        # The 'test_prefix' part of the path should have been removed
        self.assertTrue(os.path.exists(os.path.join("dst", "test.txt")))

    def test_strip_common_prefix_symlink(self):
        # Create tar file for testing
        os.makedirs(os.path.join("src", "test_prefix"))
        file_to_tar = os.path.join("src", "test_prefix", "test.txt")
        open(file_to_tar, "w").close()

        file_to_link = os.path.join("src", "test_prefix", "link.txt")
        os.symlink("./test.txt", file_to_link)
        self.assertTrue(os.path.islink(file_to_link))

        def check_for_symlink(tarinfo):
            self.assertTrue(tarinfo.issym())
            self.assertThat(file_to_link, Equals(tarinfo.name))
            self.assertThat(
                file_to_tar,
                Equals(
                    os.path.normpath(
                        os.path.join(os.path.dirname(file_to_tar), tarinfo.linkname)
                    )
                ),
            )
            return tarinfo

        tar = tarfile.open(os.path.join("src", "test.tar"), "w")
        tar.add(file_to_tar)
        tar.add(file_to_link, filter=check_for_symlink)
        tar.close()

        tar_source = sources.Tar(os.path.join("src", "test.tar"), "dst")
        os.mkdir("dst")
        tar_source.pull()

        # The 'test_prefix' part of the path should have been removed
        self.assertTrue(os.path.exists(os.path.join("dst", "test.txt")))
        self.assertTrue(os.path.exists(os.path.join("dst", "link.txt")))

    def test_strip_common_prefix_hardlink(self):
        # Create tar file for testing
        os.makedirs(os.path.join("src", "test_prefix"))
        file_to_tar = os.path.join("src", "test_prefix", "test.txt")
        open(file_to_tar, "w").close()

        file_to_link = os.path.join("src", "test_prefix", "link.txt")
        os.link(file_to_tar, file_to_link)
        self.assertTrue(os.path.exists(file_to_link))

        def check_for_hardlink(tarinfo):
            self.assertTrue(tarinfo.islnk())
            self.assertFalse(tarinfo.issym())
            self.assertThat(file_to_link, Equals(tarinfo.name))
            self.assertThat(file_to_tar, Equals(tarinfo.linkname))
            return tarinfo

        tar = tarfile.open(os.path.join("src", "test.tar"), "w")
        tar.add(file_to_tar)
        tar.add(file_to_link, filter=check_for_hardlink)
        tar.close()

        tar_source = sources.Tar(os.path.join("src", "test.tar"), "dst")
        os.mkdir("dst")
        tar_source.pull()

        # The 'test_prefix' part of the path should have been removed
        self.assertTrue(os.path.exists(os.path.join("dst", "test.txt")))
        self.assertTrue(os.path.exists(os.path.join("dst", "link.txt")))

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["tar"] is sources.Tar)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import fixtures
import jsonschema
import pytest
from testtools.matchers import Contains, DirExists, Equals, HasLength, Not

from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import go
from snapcraft.project import Project
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


class GoPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        def fake_go_build(command, cwd, *args, **kwargs):
            if command[0] == "go" and command[1] == "build" and "-o" in command:
                output = command[command.index("-o") + 1]
                if os.path.basename(output) != "module":
                    output = os.path.join(output, "binary")
                open(output, "w").close()
            elif command[0] == "go" and command[1] == "build" and "-o" not in command:
                # the package is -1
                open(os.path.join(cwd, os.path.basename(command[-1])), "w").close()

        fake_run = self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.common.run", side_effect=fake_go_build
            )
        )
        self.run_mock = fake_run.mock

        fake_run_output = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.common.run_output")
        )
        self.run_output_mock = fake_run_output.mock


class GoPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = go.GoPlugin.schema()

        properties = schema["properties"]
        for expected in ["go-packages", "go-importpath", "go-buildtags"]:
            self.assertTrue(
                expected in properties,
                "Expected {!r} to be included in properties".format(expected),
            )

        # Check go-channel
        go_channel = properties["go-channel"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in go_channel,
                "Expected {!r} to be included in 'go-channel'".format(expected),
            )

        go_channel_type = go_channel["type"]
        self.assertThat(
            go_channel_type,
            Equals("string"),
            'Expected "go-channel" "type" to be "string", but '
            'it was "{}"'.format(go_channel_type),
        )

        go_channel_default = go_channel["default"]
        self.assertThat(
            go_channel_default,
            Equals("latest/stable"),
            'Expected "go-channel" "default" to be '
            '"latest/stable", but it was "{}"'.format(go_channel_default),
        )

        # Check go-packages
        go_packages = properties["go-packages"]
        for expected in ["type", "default", "minitems", "uniqueItems", "items"]:
            self.assertTrue(
                expected in go_packages,
                "Expected {!r} to be included in 'go-packages'".format(expected),
            )

        go_packages_type = go_packages["type"]
        self.assertThat(
            go_packages_type,
            Equals("array"),
            'Expected "go-packages" "type" to be "array", but '
            'it was "{}"'.format(go_packages_type),
        )

        go_packages_default = go_packages["default"]
        self.assertThat(
            go_packages_default,
            Equals([]),
            'Expected "go-packages" "default" to be '
            '"d[]", but it was "{}"'.format(go_packages_default),
        )

        go_packages_minitems = go_packages["minitems"]
        self.assertThat(
            go_packages_minitems,
            Equals(1),
            'Expected "go-packages" "minitems" to be 1, but '
            "it was {}".format(go_packages_minitems),
        )

        self.assertTrue(go_packages["uniqueItems"])

        go_packages_items = go_packages["items"]
        self.assertTrue(
            "type" in go_packages_items,
            'Expected "type" to be included in "go-packages" ' '"items"',
        )

        go_packages_items_type = go_packages_items["type"]
        self.assertThat(
            go_packages_items_type,
            Equals("string"),
            'Expected "go-packages" "item" "type" to be '
            '"string", but it was "{}"'.format(go_packages_items_type),
        )

        # Check go-importpath
        go_importpath = properties["go-importpath"]
        for expected in ["type"]:
            self.assertTrue(
                expected in go_importpath,
                "Expected {!r} to be included in 'go-importpath'".format(expected),
            )

        go_importpath_type = go_importpath["type"]
        self.assertThat(
            go_importpath_type,
            Equals("string"),
            'Expected "go-importpath" "type" to be "string", but '
            'it was "{}"'.format(go_importpath_type),
        )

        go_importpath_default = go_importpath["default"]
        self.assertThat(
            go_importpath_default,
            Equals(""),
            'Expected "go-default" "default" to be "'
            '", but '
            'it was "{}"'.format(go_importpath_default),
        )

        # Check go-buildtags
        go_buildtags = properties["go-buildtags"]
        for expected in ["type", "default", "minitems", "uniqueItems", "items"]:
            self.assertTrue(
                expected in go_buildtags,
                "Expected {!r} to be included in 'go-buildtags'".format(expected),
            )

        go_buildtags_type = go_buildtags["type"]
        self.assertThat(
            go_buildtags_type,
            Equals("array"),
            'Expected "go-buildtags" "type" to be "array", but '
            'it was "{}"'.format(go_buildtags_type),
        )

        go_buildtags_default = go_buildtags["default"]
        self.assertThat(
            go_buildtags_default,
            Equals([]),
            'Expected "go-buildtags" "default" to be "[]", but '
            'it was "{}"'.format(go_buildtags_type),
        )

        go_buildtags_minitems = go_buildtags["minitems"]
        self.assertThat(
            go_buildtags_minitems,
            Equals(1),
            'Expected "go-buildtags" "minitems" to be 1, but '
            "it was {}".format(go_buildtags_minitems),
        )

        self.assertTrue(go_buildtags["uniqueItems"])

        go_buildtags_items = go_buildtags["items"]
        self.assertTrue(
            "type" in go_buildtags_items,
            'Expected "type" to be included in "go-buildtags" ' '"items"',
        )

        go_buildtags_items_type = go_buildtags_items["type"]
        self.assertThat(
            go_buildtags_items_type,
            Equals("string"),
            'Expected "go-buildtags" "item" "type" to be '
            '"string", but it was "{}"'.format(go_packages_items_type),
        )

        # Check required properties
        self.assertNotIn("required", schema)

    def test_get_pull_properties(self):
        expected_pull_properties = ["go-packages", "go-channel"]
        resulting_pull_properties = go.GoPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["go-packages", "go-buildtags", "go-channel"]
        resulting_build_properties = go.GoPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


class MockElfFile:
    def __init__(self, *, path: str) -> None:
        self.path = path
        self.is_dynamic = True


class GoPluginTest(GoPluginBaseTest):
    def setUp(self):
        super().setUp()

        self.useFixture(fixture_setup.CleanEnvironment())

        patcher = mock.patch("sys.stdout")
        patcher.start()
        self.addCleanup(patcher.stop)

    def assert_go_paths(self, plugin: go.GoPlugin) -> None:
        self.assertThat(plugin._gopath, DirExists())
        self.assertThat(plugin._gopath_src, DirExists())

    def test_pull_local_sources(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "main.go"), "w").close()

        plugin.pull()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "get", "-t", "-d", "./dir/..."],
                    cwd=plugin._gopath_src,
                    env=mock.ANY,
                )
            ]
        )

        self.assert_go_paths(plugin)

    def test_pull_go_mod(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""

        self.run_output_mock.return_value = "go version go1.13 linux/amd64"

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "go.mod"), "w").close()

        plugin.pull()

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.run_mock.assert_called_once_with(
            ["go", "mod", "download"], cwd=plugin.sourcedir, env=mock.ANY
        )

    def test_pull_go_mod_with_older_go_version(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = "foo"
            go_buildtags = ""

        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        self.run_output_mock.return_value = "go version go1.11 linux/amd64"

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "go.mod"), "w").close()

        plugin.pull()

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.run_mock.assert_called_once_with(
            ["go", "mod", "download"], cwd=plugin.sourcedir, env=mock.ANY
        )

        # Call pull again and ensure nothing new is logged.
        plugin.pull()

        self.assertThat(
            fake_logger.output.strip(),
            Equals(
                "Ensure build environment configuration is correct for this version of Go. "
                "Read more about it at "
                "https://github.com/golang/go/wiki/Modules#how-to-install-and-activate-module-support"
            ),
        )

    def test_go_mod_requires_newer_go_version(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        self.run_output_mock.return_value = "go version go1.6.4 linux/amd64"

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "go.mod"), "w").close()

        self.assertRaises(go.GoModRequiredVersionError, plugin.pull)

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.run_mock.assert_not_called()

    def test_no_local_source_with_go_packages(self):
        class Options:
            source = None
            go_channel = "latest/stable"
            go_packages = ["github.com/gotools/vet"]
            go_importpath = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)

        plugin.pull()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "get", "-t", "-d", plugin.options.go_packages[0]],
                    env=mock.ANY,
                    cwd=plugin._gopath_src,
                )
            ]
        )

        self.assert_go_paths(plugin)

    def test_pull_with_local_sources_or_go_packages(self):
        class Options:
            source = None
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)
        plugin.pull()

        self.run_mock.assert_has_calls([])

        self.assert_go_paths(plugin)

    def test_build_with_local_sources(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "main.go"), "w").close()

        plugin.pull()

        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()
        self.run_output_mock.reset_mock()
        self.run_output_mock.return_value = "dir/pkg/main main"

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            ["go", "list", "-f", "{{.ImportPath}} {{.Name}}", "./dir/..."],
            cwd=plugin._gopath_src,
            env=mock.ANY,
        )

        self.run_mock.assert_called_once_with(
            ["go", "build", "dir/pkg/main"],
            cwd=os.path.join(plugin.installdir, "bin"),
            env=mock.ANY,
        )

        self.assert_go_paths(plugin)

    def test_build_go_packages(self):
        class Options:
            source = ""
            go_channel = "latest/stable"
            go_packages = ["github.com/gotools/vet"]
            go_importpath = ""
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)

        plugin.pull()

        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()
        plugin.build()

        self.run_mock.assert_called_once_with(
            ["go", "build", plugin.options.go_packages[0]],
            cwd=os.path.join(plugin.installdir, "bin"),
            env=mock.ANY,
        )

        self.assert_go_paths(plugin)

    def test_clean_build(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        plugin.pull()

        os.makedirs(plugin._gopath_pkg)
        os.makedirs(plugin.builddir)

        plugin.build()

        self.assert_go_paths(plugin)

        plugin.clean_build()

        self.assertTrue(os.path.exists(plugin._gopath))
        self.assertTrue(os.path.exists(plugin._gopath_src))
        self.assertFalse(os.path.exists(plugin._gopath_bin))
        self.assertFalse(os.path.exists(plugin._gopath_pkg))

    def test_clean_pull(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "main.go"), "w").close()

        plugin.pull()

        self.assertTrue(os.path.exists(plugin._gopath))

        plugin.clean_pull()

        self.assertFalse(os.path.exists(plugin._gopath))

    def test_build_with_local_sources_and_go_importpath(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = "github.com/snapcore/launcher"
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "main.go"), "w").close()

        plugin.pull()

        os.makedirs(plugin.builddir)
        self.run_output_mock.return_value = "github.com/snapcore/launcher main"

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            [
                "go",
                "list",
                "-f",
                "{{.ImportPath}} {{.Name}}",
                "./github.com/snapcore/launcher/...",
            ],
            cwd=plugin._gopath_src,
            env=mock.ANY,
        )

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "get", "-t", "-d", "./github.com/snapcore/launcher/..."],
                    cwd=plugin._gopath_src,
                    env=mock.ANY,
                ),
                mock.call(
                    ["go", "build", "github.com/snapcore/launcher"],
                    cwd=os.path.join(plugin.installdir, "bin"),
                    env=mock.ANY,
                ),
            ]
        )

        self.assertTrue(
            os.path.exists(
                os.path.join(plugin._gopath_src, plugin.options.go_importpath)
            )
        )

        self.assert_go_paths(plugin)

    def test_build_environment(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)
        open(os.path.join(plugin.sourcedir, "main.go"), "w").close()
        os.makedirs(os.path.join(plugin.installdir, "lib"))
        os.makedirs(os.path.join(plugin.installdir, "usr", "lib"))
        os.makedirs(os.path.join(plugin.project.stage_dir, "lib"))
        os.makedirs(os.path.join(plugin.project.stage_dir, "usr", "lib"))
        plugin.pull()

        self.assertThat(self.run_mock.call_count, Equals(1))
        for call_args in self.run_mock.call_args_list:
            env = call_args[1]["env"]
            self.assertTrue("GOPATH" in env, "Expected environment to include GOPATH")
            self.assertThat(env["GOPATH"], Equals(plugin._gopath))

            self.assertTrue(
                "CGO_LDFLAGS" in env, "Expected environment to include CGO_LDFLAGS"
            )
            expected_flags = [
                "-L{}/lib".format(plugin.installdir),
                "-L{}/usr/lib".format(plugin.installdir),
                "-L{}/lib".format(plugin.project.stage_dir),
                "-L{}/usr/lib".format(plugin.project.stage_dir),
            ]
            for flag in expected_flags:
                self.assertTrue(
                    flag in env["CGO_LDFLAGS"],
                    "Expected $CGO_LDFLAGS to include {!r}, but it was "
                    '"{}"'.format(flag, env["CGO_LDFLAGS"]),
                )

    def test_build_with_buildtag(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_importpath = ""
            go_packages = []
            go_buildtags = ["testbuildtag1", "testbuildtag2"]

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.options.source)
        os.makedirs(plugin.sourcedir)

        plugin.pull()

        os.makedirs(plugin._gopath_bin)
        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()
        self.run_output_mock.return_value = "dir/pkg/main main"

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            ["go", "list", "-f", "{{.ImportPath}} {{.Name}}", "./dir/..."],
            cwd=plugin._gopath_src,
            env=mock.ANY,
        )

        self.run_mock.assert_called_once_with(
            ["go", "build", "-tags=testbuildtag1,testbuildtag2", "dir/pkg/main"],
            cwd=os.path.join(plugin.installdir, "bin"),
            env=mock.ANY,
        )

    def test_build_go_mod(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        self.run_output_mock.return_value = "go version go1.13 linux/amd64"

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.builddir)
        open(os.path.join(plugin.builddir, "go.mod"), "w").close()

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.run_mock.assert_called_once_with(
            ["go", "build", "-o", plugin._install_bin_dir, "./..."],
            cwd=plugin.builddir,
            env=mock.ANY,
        )

    def test_build_go_mod_with_older_go_version(self):
        class Options:
            source = "dir"
            go_channel = "latest/stable"
            go_packages = []
            go_importpath = ""
            go_buildtags = ""

        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        self.run_output_mock.return_value = "go version go1.11 linux/amd64"

        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.builddir)
        with open(os.path.join(plugin.builddir, "go.mod"), "w") as go_mod_file:
            print("module github.com/foo/bar/module", file=go_mod_file)
            print("go 1.11", file=go_mod_file)

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.run_mock.assert_called_once_with(
            ["go", "build", "-o", f"{plugin._install_bin_dir}/module"],
            cwd=plugin.builddir,
            env=mock.ANY,
        )

        self.assertThat(
            fake_logger.output.strip(),
            Equals(
                "Ensure build environment configuration is correct for this version of Go. "
                "Read more about it at "
                "https://github.com/golang/go/wiki/Modules#how-to-install-and-activate-module-support"
            ),
        )

    @mock.patch("snapcraft.internal.elf.ElfFile")
    def test_build_classic_dynamic_relink(self, mock_elffile):
        class Options:
            source = ""
            go_channel = "latest/stable"
            go_packages = ["github.com/gotools/vet"]
            go_importpath = ""
            go_buildtags = ""

        mock_elffile.return_value = MockElfFile(path="foo")
        self.project._snap_meta.confinement = "classic"
        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.sourcedir)

        plugin.pull()

        os.makedirs(plugin.builddir)

        self.run_mock.reset_mock()
        plugin.build()

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "build", plugin.options.go_packages[0]],
                    cwd=os.path.join(plugin.installdir, "bin"),
                    env=mock.ANY,
                ),
                mock.call(
                    [
                        "go",
                        "build",
                        "-ldflags",
                        "-linkmode=external",
                        plugin.options.go_packages[0],
                    ],
                    cwd=os.path.join(plugin.installdir, "bin"),
                    env=mock.ANY,
                ),
            ]
        )

        self.assert_go_paths(plugin)

    @mock.patch("snapcraft.internal.elf.ElfFile")
    def test_build_go_mod_classic_dynamic_relink(self, mock_elffile):
        class Options:
            source = ""
            go_channel = "latest/stable"
            go_packages = ["github.com/gotools/vet"]
            go_importpath = ""
            go_buildtags = ""

        self.run_output_mock.return_value = "go version go13 linux/amd64"

        mock_elffile.return_value = MockElfFile(path="foo")
        self.project._snap_meta.confinement = "classic"
        plugin = go.GoPlugin("test-part", Options(), self.project)

        os.makedirs(plugin.builddir)
        open(os.path.join(plugin.builddir, "go.mod"), "w").close()

        plugin.build()

        self.run_output_mock.assert_called_once_with(
            ["go", "version"], cwd=mock.ANY, env=mock.ANY
        )
        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["go", "build", "-o", plugin._install_bin_dir, "./..."],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(
                    [
                        "go",
                        "build",
                        "-ldflags",
                        "-linkmode=external",
                        "-o",
                        plugin._install_bin_dir,
                        "./...",
                    ],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )


class GoPluginSchemaValidationTest(unit.TestCase):
    def test_sources_validation_neither(self):
        schema = self._get_schema()
        properties = {}
        self.assertRaises(
            jsonschema.ValidationError, jsonschema.validate, properties, schema
        )

    def test_sources_validation_source(self):
        schema = self._get_schema()
        properties = {"source": ""}
        jsonschema.validate(properties, schema)

    def test_sources_validation_packages(self):
        schema = self._get_schema()
        properties = {"go-packages": []}
        jsonschema.validate(properties, schema)

    def test_sources_validation_both(self):
        schema = self._get_schema()
        properties = {"source": "foo", "go-packages": []}
        jsonschema.validate(properties, schema)

    def _get_schema(self):
        schema = go.GoPlugin.schema()
        # source definition comes from the main schema
        schema["properties"]["source"] = {"type": "string"}
        return schema


class GoPluginToolSetupTest(GoPluginBaseTest):
    def setUp(self):
        super().setUp()

        class Options:
            source = "dir"
            go_channel = "latest/stable"

        self.options = Options()

    def test_snap(self):
        plugin = go.GoPlugin("test-part", self.options, self.project)

        self.assertThat(plugin.build_packages, Not(Contains("golang-go")))
        self.assertThat(plugin.build_snaps, Contains("go/latest/stable"))

    def test_build_packages(self):
        self.options.go_channel = ""

        plugin = go.GoPlugin("test-part", self.options, self.project)

        self.assertThat(plugin.build_packages, Contains("golang-go"))
        self.assertThat(plugin.build_snaps, Not(Contains("go/latest/stable")))


class GoPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"
            go_channel = "latest/stable"

        self.options = Options()

    def test_unsupported_base_using_snap(self):
        self.assertRaises(
            errors.PluginBaseError, go.GoPlugin, "test-part", self.options, self.project
        )

    def test_unsupported_base_using_without_snap_raises(self):
        self.options.go_channel = ""

        self.assertRaises(
            errors.PluginBaseError, go.GoPlugin, "test-part", self.options, self.project
        )


@pytest.mark.parametrize(
    "deb_arch,go_arch",
    [
        ("armhf", "arm"),
        ("arm64", "arm64"),
        ("i386", "386"),
        ("amd64", "amd64"),
        ("ppc64el", "ppc64le"),
    ],
)
def test_cross_compile(monkeypatch, tmp_work_path, mock_run, deb_arch, go_arch):
    monkeypatch.setattr(Project, "is_cross_compiling", True)

    class Options:
        source = ""
        go_packages = ["github.com/gotools/vet"]
        go_importpath = ""
        go_buildtags = ""
        go_channel = ""

    project = Project(target_deb_arch=deb_arch)
    project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

    plugin = go.GoPlugin("test-part", Options(), project)

    os.makedirs(plugin.sourcedir)

    plugin.pull()

    assert mock_run.call_count == 1

    for call_args in mock_run.call_args_list:
        env = call_args[1]["env"]
        assert "CC" in env
        assert env["CC"] == f"{project.arch_triplet}-gcc"

        assert "CXX" in env
        assert env["CXX"] == f"{project.arch_triplet}-g++"

        assert "CGO_ENABLED" in env
        assert env["CGO_ENABLED"] == "1"

        assert "GOARCH" in env
        assert env["GOARCH"] == go_arch

        if deb_arch == "armhf":
            assert "GOARM" in env
            assert env["GOARM"] == "7"


class TestCGoLdFlags:
    scenarios = (
        (
            "none",
            dict(cgo_ldflags_env="", library_paths=[], ldflags_env="", expected=""),
        ),
        (
            "CGO_LDFLAGS",
            dict(
                cgo_ldflags_env="-lbar",
                library_paths=[],
                ldflags_env="",
                expected="-lbar",
            ),
        ),
        (
            "Library Paths",
            dict(
                cgo_ldflags_env="",
                library_paths=["part/part1/usr/lib", "stage/lib"],
                ldflags_env="",
                expected="-Lpart/part1/usr/lib -Lstage/lib",
            ),
        ),
        (
            "LDFLAGS",
            dict(
                cgo_ldflags_env="",
                library_paths=[],
                ldflags_env="-lfoo",
                expected="-lfoo",
            ),
        ),
        (
            "all",
            dict(
                cgo_ldflags_env="-lbar",
                library_paths=["part/part1/usr/lib", "stage/lib"],
                ldflags_env="-lfoo",
                expected="-lbar -Lpart/part1/usr/lib -Lstage/lib -lfoo",
            ),
        ),
    )

    def test_generated_cgo_ldflags(
        self, monkeypatch, cgo_ldflags_env, library_paths, ldflags_env, expected
    ):
        monkeypatch.setenv("CGO_LDFLAGS", cgo_ldflags_env)
        monkeypatch.setenv("LDFLAGS", ldflags_env)

        assert go._get_cgo_ldflags(library_paths) == expected



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from textwrap import dedent

import testscenarios
from testtools.matchers import Contains, FileContains, FileExists, Not

from tests import integration


class PrimeTestCase(integration.TestCase):
    def test_prime_includes_stage_fileset(self):
        self.run_snapcraft("prime", "prime-from-stage")
        self.assertThat(os.path.join(self.prime_dir, "without-a"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "without-b"), Not(FileExists()))
        self.assertThat(os.path.join(self.prime_dir, "without-c"), FileExists())

    def test_prime_includes_stage_excludes_fileset(self):
        self.run_snapcraft("prime", "prime-from-stage")
        self.assertThat(os.path.join(self.prime_dir, "with-a"), Not(FileExists()))
        self.assertThat(os.path.join(self.prime_dir, "with-b"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "with-c"), FileExists())

    def test_prime_with_non_ascii_desktop_file(self):
        # Originally, in this test we forced LC_ALL=C. However, now that we
        # are using the click python library we can't do it because it fails
        # to run any command when the system language is ascii.
        # --20170518 - elopio
        self.run_snapcraft("prime", "desktop-with-non-ascii")

        desktop_path = os.path.join(self.prime_dir, "meta", "gui", "test-app.desktop")

        self.expectThat(desktop_path, FileContains(matcher=Contains("non ascíí")))


class PrimedAssetsTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        ("setup", dict(project_dir="assets-with-gui-in-setup")),
        ("snap", dict(project_dir="assets-with-gui-in-snap")),
    ]

    def test_assets_in_meta(self):
        self.run_snapcraft("prime", self.project_dir)

        gui_dir = os.path.join(self.prime_dir, "meta", "gui")
        expected_desktop = dedent(
            """\
            [Desktop Entry]
            Name=My App
            Exec=my-app
            Type=Application
            """
        )
        self.expectThat(os.path.join(gui_dir, "icon.png"), FileExists())
        self.expectThat(
            os.path.join(gui_dir, "my-app.desktop"), FileContains(expected_desktop)
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
import csv
import logging
from typing import Dict, List  # noqa: F401

from snapcraft.internal import errors

logger = logging.getLogger(__name__)


class Mount:
    """A class to provide programmatic access to a specific mountpoint"""

    def __init__(self, mountinfo_row: List[str]) -> None:
        # Parse the row according to section 3.5 of
        # https://www.kernel.org/doc/Documentation/filesystems/proc.txt
        try:
            self.mount_id = mountinfo_row[0]
            self.parent_id = mountinfo_row[1]
            self.st_dev = mountinfo_row[2]
            self.root = mountinfo_row[3]
            self.mount_point = mountinfo_row[4]
            self.mount_options = mountinfo_row[5]
            separator_index = mountinfo_row.index("-")
            self.optional_fields = mountinfo_row[6:separator_index]
            self.filesystem_type = mountinfo_row[separator_index + 1]
            self.mount_source = mountinfo_row[separator_index + 2]
            self.super_options = mountinfo_row[separator_index + 3]
        except IndexError as e:
            raise errors.InvalidMountinfoFormat(" ".join(mountinfo_row)) from e


class MountInfo:
    """A class to provide programmatic access to /proc/self/mountinfo"""

    def __init__(self, *, mountinfo_file: str = "/proc/self/mountinfo") -> None:
        """Create a new MountInfo instance.

        :param str mountinfo_file: Path to mountinfo file to be parsed.
        """
        # Maintain two dicts pointing to the same underlying objects:
        # a dict of mount points to Mounts, and a dict of roots to Mounts.
        self._mount_point_mounts = {}  # type: Dict[str, Mount]
        root_mounts = collections.defaultdict(
            list
        )  # type: Dict[str, List[Mount]]  # noqa

        with contextlib.suppress(FileNotFoundError):
            with open(mountinfo_file) as f:
                for row in csv.reader(f, delimiter=" "):
                    try:
                        mount = Mount(row)
                        self._mount_point_mounts[mount.mount_point] = mount
                        root_mounts[mount.root].append(mount)
                    except errors.InvalidMountinfoFormat as e:
                        logger.warning(str(e))

        self._root_mounts = dict(root_mounts)

    def for_mount_point(self, mount_point: str) -> Mount:
        try:
            return self._mount_point_mounts[mount_point]
        except KeyError:
            raise errors.MountPointNotFoundError(mount_point)

    def for_root(self, root: str) -> List[Mount]:
        try:
            return self._root_mounts[root]
        except KeyError:
            raise errors.RootNotMountedError(root)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The make plugin is useful for building make based parts.

Make based projects are projects that have a Makefile that drives the
build.

This plugin always runs 'make' followed by 'make install', except when
the 'artifacts' keyword is used.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - artifacts:
      (list)
      Link/copy the given files from the make output to the snap
      installation directory. If specified, the 'make install'
      step will be skipped.

    - makefile:
      (string)
      Use the given file as the makefile.

    - make-parameters:
      (list of strings)
      Pass the given parameters to the make command.

    - make-install-var:
      (string; default: DESTDIR)
      Use this variable to redirect the installation into the snap.
"""

import os

import snapcraft.common
from snapcraft.plugins.v1 import PluginV1


class MakePlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["makefile"] = {"type": "string"}
        schema["properties"]["make-parameters"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["make-install-var"] = {
            "type": "string",
            "default": "DESTDIR",
        }
        schema["properties"]["artifacts"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["makefile", "make-parameters", "make-install-var", "artifacts"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.build_packages.append("make")

    def make(self, env=None):
        command = ["make"]

        if self.options.makefile:
            command.extend(["-f", self.options.makefile])

        self.run(
            command + [f"-j{self.parallel_build_count}"] + self.options.make_parameters,
            env=env,
        )
        if self.options.artifacts:
            for artifact in self.options.artifacts:
                source_path = os.path.join(self.builddir, artifact)
                destination_path = os.path.join(self.installdir, artifact)
                if os.path.isdir(source_path):
                    snapcraft.file_utils.link_or_copy_tree(
                        source_path, destination_path
                    )
                else:
                    snapcraft.file_utils.link_or_copy(source_path, destination_path)
        else:
            install_command = command + ["install"] + self.options.make_parameters
            if self.options.make_install_var:
                install_command.append(
                    "{}={}".format(self.options.make_install_var, self.installdir)
                )

            self.run(install_command, env=env)

    def build(self):
        super().build()
        self.make()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import pytest

from snapcraft.plugins.v1 import catkin_tools

from . import PluginsV1BaseTestCase


class CatkinToolsPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class props:
            rosdistro = "indigo"
            ubuntu_distro = "trusty"
            catkin_packages = ["my_package"]
            source_space = "src"
            source_subdir = None
            include_roscore = False
            catkin_cmake_args = []
            underlay = None
            rosinstall_files = None
            build_attributes = []

        self.properties = props()

        patcher = mock.patch("snapcraft.plugins.v1._python.Pip")
        self.pip_mock = patcher.start()
        self.addCleanup(patcher.stop)
        self.pip_mock.return_value.list.return_value = {}


class CatkinToolsPluginTestCase(CatkinToolsPluginBaseTest):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.repo.Ubuntu")
        self.ubuntu_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "run")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "_prepare_build")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "_finish_build")
    def test_build_multiple(
        self, finish_build_mock, prepare_build_mock, run_output_mock, run_mock
    ):
        self.properties.catkin_packages.append("package_2")

        plugin = catkin_tools.CatkinToolsPlugin(
            "test-part", self.properties, self.project
        )
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        class check_pkg_arguments:
            def __init__(self, test):
                self.test = test

            def __eq__(self, args):
                index = args.index("build")
                packages = args[index + 1 :]
                self.test.assertIn("my_package", packages)
                self.test.assertIn("package_2", packages)
                return True

        run_mock.assert_called_with(check_pkg_arguments(self))

        finish_build_mock.assert_called_once_with()

    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "run")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "run_output", return_value="foo")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "_prepare_build")
    @mock.patch.object(catkin_tools.CatkinToolsPlugin, "_finish_build")
    def test_build(
        self, finish_build_mock, prepare_build_mock, run_output_mock, run_mock
    ):
        plugin = catkin_tools.CatkinToolsPlugin(
            "test-part", self.properties, self.project
        )
        os.makedirs(os.path.join(plugin.sourcedir, "src"))

        plugin.build()

        prepare_build_mock.assert_called_once_with()

        class check_build_command:
            def __eq__(self, args):
                command = " ".join(args)
                return (
                    args[0] == "catkin"
                    and "build" in command
                    and "my_package" in command
                )

        run_mock.assert_called_with(check_build_command())

        finish_build_mock.assert_called_once_with()


@pytest.fixture
def options():
    class Options:
        rosdistro = "indigo"
        ubuntu_distro = "trusty"
        catkin_packages = ["my_package"]
        source_space = "src"
        source_subdir = None
        include_roscore = False
        catkin_cmake_args = []
        underlay = None
        rosinstall_files = None
        build_attributes = []

    return Options()


class TestPrepareBuild:

    scenarios = [
        (
            "release without catkin-cmake-args",
            {"build_attributes": [], "catkin_cmake_args": []},
        ),
        (
            "release with catkin-cmake-args",
            {"build_attributes": [], "catkin_cmake_args": ["-DFOO"]},
        ),
        (
            "debug without catkin-cmake-args",
            {"build_attributes": ["debug"], "catkin_cmake_args": []},
        ),
        (
            "debug with catkin-cmake-args",
            {"build_attributes": ["debug"], "catkin_cmake_args": ["-DFOO"]},
        ),
    ]

    def test_prepare_build(
        self,
        monkeypatch,
        mock_run,
        project,
        options,
        build_attributes,
        catkin_cmake_args,
    ):
        monkeypatch.setattr(
            catkin_tools.CatkinToolsPlugin, "_use_in_snap_python", lambda x: True
        )

        options.build_attributes.extend(build_attributes)
        options.catkin_cmake_args = catkin_cmake_args

        plugin = catkin_tools.CatkinToolsPlugin("test-part", options, project)
        os.makedirs(os.path.join(plugin.rosdir, "test"))

        plugin._prepare_build()

        if "debug" in options.build_attributes:
            expected_cmake_args = ["-DCMAKE_BUILD_TYPE=Debug"] + catkin_cmake_args
        else:
            expected_cmake_args = ["-DCMAKE_BUILD_TYPE=Release"] + catkin_cmake_args

        mock_run.assert_has_calls(
            [
                mock.call(["catkin", "init"]),
                mock.call(["catkin", "clean", "-y"]),
                mock.call(["catkin", "profile", "add", "-f", "default"]),
                mock.call(
                    [
                        "catkin",
                        "config",
                        "--profile",
                        "default",
                        "--build-space",
                        plugin.builddir,
                        "--source-space",
                        os.path.join(plugin.builddir, plugin.options.source_space),
                        "--install",
                        "--install-space",
                        plugin.rosdir,
                        "--cmake-args",
                    ]
                    + expected_cmake_args
                ),
            ]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2017-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import cmake
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


class CMakeBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            configflags = []
            source_subdir = None

            make_parameters = []
            disable_parallel = False
            build_snaps = []

        self.options = Options()

        patcher = mock.patch("snapcraft.internal.common.run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.useFixture(fixture_setup.CleanEnvironment())


class CMakeTest(CMakeBaseTest):
    def test_get_build_properties(self):
        expected_build_properties = ["configflags"]
        resulting_build_properties = cmake.CMakePlugin.get_build_properties()
        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    def test_build_referencing_sourcedir_if_no_subdir(self):
        plugin = cmake.CMakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["cmake", plugin.sourcedir, "-DCMAKE_INSTALL_PREFIX="],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["cmake", "--build", ".", "--", "-j2"],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["cmake", "--build", ".", "--target", "install"],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_referencing_sourcedir_with_subdir(self):
        self.options.source_subdir = "subdir"

        plugin = cmake.CMakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)
        plugin.build()

        sourcedir = os.path.join(plugin.sourcedir, plugin.options.source_subdir)
        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["cmake", sourcedir, "-DCMAKE_INSTALL_PREFIX="],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["cmake", "--build", ".", "--", "-j2"],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["cmake", "--build", ".", "--target", "install"],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_build_disable_parallel(self):
        self.options.disable_parallel = True

        plugin = cmake.CMakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)
        plugin.build()

        self.run_mock.assert_has_calls(
            [
                mock.call(
                    ["cmake", "--build", ".", "--", "-j1"],
                    cwd=plugin.builddir,
                    env=mock.ANY,
                )
            ]
        )

    def test_build_environment(self):
        plugin = cmake.CMakePlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)
        plugin.build()

        expected = {}

        expected["DESTDIR"] = plugin.installdir
        expected["CMAKE_PREFIX_PATH"] = "$CMAKE_PREFIX_PATH:{}".format(self.stage_dir)
        expected["CMAKE_INCLUDE_PATH"] = "$CMAKE_INCLUDE_PATH:" + ":".join(
            ["{0}/include", "{0}/usr/include", "{0}/include/{1}", "{0}/usr/include/{1}"]
        ).format(self.stage_dir, plugin.project.arch_triplet)
        expected["CMAKE_LIBRARY_PATH"] = "$CMAKE_LIBRARY_PATH:" + ":".join(
            ["{0}/lib", "{0}/usr/lib", "{0}/lib/{1}", "{0}/usr/lib/{1}"]
        ).format(self.stage_dir, plugin.project.arch_triplet)

        self.assertThat(self.run_mock.call_count, Equals(3))
        for call_args in self.run_mock.call_args_list:
            environment = call_args[1]["env"]
            for variable, value in expected.items():
                self.assertTrue(
                    variable in environment,
                    'Expected variable "{}" to be in environment'.format(variable),
                )

                self.assertThat(
                    environment[variable],
                    Equals(value),
                    "Expected ${}={}, but it was {}".format(
                        variable, value, environment[variable]
                    ),
                )

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            cmake.CMakePlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))


class TestSnapCMakeBuild:

    scenarios = [
        ("no snaps", dict(build_snaps=[], expected_config_flags=[])),
        (
            "one build snap",
            dict(
                snaps=["kde-plasma-sdk"],
                expected_configflags=[
                    "-DCMAKE_FIND_ROOT_PATH=/snap/kde-plasma-sdk/current"
                ],
            ),
        ),
        (
            "one build snap, preexisting root find path",
            dict(
                snaps=["kde-plasma-sdk"],
                expected_configflags=[
                    "-DCMAKE_FIND_ROOT_PATH=/snap/kde-plasma-sdk/current"
                ],
            ),
            (
                "one build snap with channel",
                dict(
                    snaps=["kde-plasma-sdk/latest/edge"],
                    expected_configflags=[
                        "-DCMAKE_FIND_ROOT_PATH=/snap/kde-plasma-sdk/current"
                    ],
                ),
            ),
        ),
        (
            "two build snap with channel",
            dict(
                snaps=["gnome-sdk", "kde-plasma-sdk/latest/edge"],
                expected_configflags=[
                    "-DCMAKE_FIND_ROOT_PATH=/snap/gnome-sdk/current;/snap/kde-plasma-sdk/current"
                ],
            ),
        ),
    ]

    def test_build(self, project, mock_run, snaps, expected_configflags):
        class Options:
            configflags = []
            source_subdir = None

            make_parameters = []
            disable_parallel = False
            build_snaps = snaps

        plugin = cmake.CMakePlugin("test-part", Options(), project)
        os.makedirs(plugin.builddir)
        plugin.build()

        mock_run.assert_has_calls(
            [
                mock.call(
                    ["cmake", plugin.sourcedir, "-DCMAKE_INSTALL_PREFIX="]
                    + expected_configflags,
                    env=mock.ANY,
                ),
                mock.call(["cmake", "--build", ".", "--", "-j2"], env=mock.ANY),
                mock.call(
                    ["cmake", "--build", ".", "--target", "install"], env=mock.ANY
                ),
            ]
        )


class FlagsTest(unit.TestCase):
    def test_simple_flag(self):
        flag = cmake._Flag("-DVERBOSE")

        self.assertThat(flag.name, Equals("-DVERBOSE"))
        self.assertThat(flag.value, Equals(None))
        self.assertThat(str(flag), Equals("-DVERBOSE"))

    def test_flag(self):
        flag = cmake._Flag("-DCMAKE_PREFIX_PATH=foo")

        self.assertThat(flag.name, Equals("-DCMAKE_PREFIX_PATH"))
        self.assertThat(flag.value, Equals("foo"))
        self.assertThat(str(flag), Equals("-DCMAKE_PREFIX_PATH=foo"))

    def test_flag_value_change(self):
        flag = cmake._Flag("-DCMAKE_PREFIX_PATH=foo")
        flag.value = "bar"

        self.assertThat(flag.name, Equals("-DCMAKE_PREFIX_PATH"))
        self.assertThat(flag.value, Equals("bar"))
        self.assertThat(str(flag), Equals("-DCMAKE_PREFIX_PATH=bar"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License remote_build 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals

import snapcraft.internal.remote_build.errors as errors
from tests import fixture_setup

from . import CommandBaseTestCase


class RemoteBuildTests(CommandBaseTestCase):
    def setUp(self):
        super().setUp()
        self.snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, parts={"part0": {"plugin": "nil"}},
        )
        self.useFixture(self.snapcraft_yaml)

        self.mock_lc = self.useFixture(
            fixtures.MockPatch("snapcraft.cli.remote.LaunchpadClient", autospec=True)
        ).mock.return_value
        self.mock_lc_architectures = mock.PropertyMock(return_value=["i386"])
        type(self.mock_lc).architectures = self.mock_lc_architectures
        self.mock_lc.has_outstanding_build.return_value = False

    def test_remote_build(self):
        result = self.run_command(["remote-build", "--launchpad-accept-public-upload"])

        self.mock_lc.start_build.assert_called_once()
        self.mock_lc.cleanup.assert_called_once()
        self.assertThat(result.output, Contains("Building snap package for i386."))
        self.assertThat(result.exit_code, Equals(0))

    def test_remote_build_multiple_arch(self):
        self.mock_lc_architectures.return_value = ["i386", "amd64", "arm64"]

        result = self.run_command(
            [
                "remote-build",
                "--launchpad-accept-public-upload",
                "--build-on",
                "i386,amd64,arm64",
            ]
        )

        self.mock_lc.start_build.assert_called_once()
        self.mock_lc.cleanup.assert_called_once()
        self.assertThat(
            result.output, Contains("Building snap package for amd64, arm64, and i386.")
        )
        self.assertThat(result.exit_code, Equals(0))

    def test_remote_build_invalid_user_arch(self):
        self.assertRaises(
            errors.UnsupportedArchitectureError,
            self.run_command,
            ["remote-build", "--launchpad-accept-public-upload", "--build-on", "x64"],
        )

        self.mock_lc.start_build.assert_not_called()
        self.mock_lc.cleanup.assert_not_called()

    @mock.patch("snapcraft.cli.remote.echo")
    def test_remote_build_sudo_errors(self, mock_echo):
        self.useFixture(fixtures.EnvironmentVariable("SUDO_USER", "testuser"))
        self.useFixture(fixtures.MockPatch("os.geteuid", return_value=0))

        self.run_command(["remote-build", "--launchpad-accept-public-upload"])
        assert mock_echo.has_calls(
            [
                mock.call.warning(
                    "Running with 'sudo' may cause permission errors and is discouraged."
                )
            ]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Dict, List, Union, cast

from ._env import environment_to_replacements, runtime_env  # noqa: F401
from ._extensions import (  # noqa: F401
    apply_extensions,
    find_extension,
    supported_extension_names,
)
from ._parts_config import PartsConfig  # noqa: F401

if TYPE_CHECKING:
    from snapcraft.project import Project  # noqa: F401


def load_config(project: "Project"):
    from ._config import Config

    return Config(project)


def replace_attr(
    attr: Union[List[str], Dict[str, str], str], replacements: Dict[str, str]
) -> Union[List[str], Dict[str, str], str]:
    if isinstance(attr, str):
        for replacement, value in replacements.items():
            attr = attr.replace(replacement, str(value))
        return attr
    elif isinstance(attr, list) or isinstance(attr, tuple):
        return [cast(str, replace_attr(i, replacements)) for i in attr]
    elif isinstance(attr, dict):
        result = dict()  # type: Dict[str, str]
        for key, value in attr.items():
            # Run replacements on both the key and value
            key = cast(str, replace_attr(key, replacements))
            value = cast(str, replace_attr(value, replacements))
            result[key] = value
        return result

    return attr



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from unittest.mock import call

import pytest

from snapcraft.internal import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.plugins.v1 import flutter
from snapcraft.project import Project


def test_schema():
    assert flutter.FlutterPlugin.schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": False,
        "properties": {
            "flutter-revision": {"default": None, "type": "string"},
            "flutter-target": {"default": "lib/main.dart", "type": "string"},
        },
        "required": ["source"],
        "type": "object",
    }


def test_get_pull_properties():
    assert flutter.FlutterPlugin.get_pull_properties() == ["flutter-revision"]


def test_get_build_properties():
    assert flutter.FlutterPlugin.get_build_properties() == ["flutter-target"]


@pytest.fixture
def flutter_options():
    """Return Options object suitable for the flutter plugin."""

    class Options:
        flutter_channel = "stable"
        flutter_target = "lib/main.dart"
        flutter_revision = None
        source_subdir = ""

    return Options()


@pytest.fixture
def flutter_plugin(tmp_work_path, project, flutter_options):
    """Return an instance of FlutterPlugin setup with different bases."""
    return flutter.FlutterPlugin("test-part", flutter_options, project)


def test_pull(mock_subprocess_run, flutter_plugin):
    flutter_plugin.pull()


def test_pull_with_revision(mock_subprocess_run, flutter_plugin):
    flutter_plugin.options.flutter_revision = "foo"
    flutter_plugin.pull()

    expected_cwd = pathlib.Path("parts/test-part/src").absolute()
    assert mock_subprocess_run.mock_calls == [
        call("yes | flutter version foo", shell=True, check=True, cwd=expected_cwd)
    ]


def test_pull_from_subdir(mock_subprocess_run, flutter_plugin):
    flutter_plugin.options.source_subdir = "subdir"
    flutter_plugin.pull()


def test_build(mock_run, flutter_plugin):
    # "mock" create the bundle directory and app.
    if flutter_plugin.project.deb_arch == "arm64":
        bundle_dir = "build/linux/arm64/release/bundle/my_app"
    else:
        bundle_dir = "build/linux/x64/release/bundle/my_app"

    app_path = pathlib.Path(flutter_plugin.builddir) / bundle_dir
    app_path.parent.mkdir(parents=True)
    app_path.touch()

    # Create the installdir.
    pathlib.Path(flutter_plugin.installdir).mkdir(parents=True)

    flutter_plugin.build()

    assert mock_run.mock_calls == [
        call(["flutter", "pub", "get"]),
        call(
            [
                "flutter",
                "build",
                "linux",
                "--release",
                "-v",
                "-t",
                flutter_plugin.options.flutter_target,
            ]
        ),
    ]
    assert (pathlib.Path(flutter_plugin.installdir) / "bin/my_app").exists()


def test_unsupported_base_raises(flutter_options):
    project = Project()
    project._snap_meta = Snap(name="test-snap", base="bad-base", confinement="strict")

    with pytest.raises(errors.PluginBaseError):
        flutter.FlutterPlugin("test-part", flutter_options, project)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal import errors


class GrammarError(errors.SnapcraftError):
    """Base class for grammar-related errors."""

    pass


class GrammarSyntaxError(GrammarError):

    fmt = "Invalid grammar syntax: {message}"

    def __init__(self, message):
        super().__init__(message=message)


class OnStatementSyntaxError(GrammarSyntaxError):
    def __init__(self, on_statement, *, message=None):
        components = ["{!r} is not a valid 'on' clause".format(on_statement)]
        if message:
            components.append(message)
        super().__init__(message=": ".join(components))


class ToStatementSyntaxError(GrammarSyntaxError):
    def __init__(self, to_statement, *, message=None):
        components = ["{!r} is not a valid 'to' clause".format(to_statement)]
        if message:
            components.append(message)
        super().__init__(message=": ".join(components))


class UnsatisfiedStatementError(GrammarError):

    fmt = "Unable to satisfy {statement!r}, failure forced"

    def __init__(self, statement):
        super().__init__(statement=statement)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from testtools.matchers import Equals

import snapcraft.internal
from snapcraft import yaml_utils
from tests import unit

from .conftest import Project


class PullStateBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.project = Project()
        self.property_names = ["foo"]
        self.part_properties = {"foo": "bar"}

        self.state = snapcraft.internal.states.PullState(
            self.property_names, self.part_properties, self.project
        )


class PullStateTestCase(PullStateBaseTestCase):
    @mock.patch.object(
        snapcraft.internal.states.PullState,
        "__init__",
        wraps=snapcraft.internal.states.PullState.__init__,
    )
    def test_yaml_conversion(self, init_spy):
        state_string = yaml_utils.dump(self.state)

        # Verify that the dumped tag was correct
        self.assertThat(state_string.splitlines()[0], Equals("!PullState"))

        # Now verify the conversion
        state_from_yaml = yaml_utils.load(state_string)
        self.assertThat(state_from_yaml, Equals(self.state))

        # Verify that init was not called
        init_spy.assert_not_called()

    def test_comparison(self):
        other = snapcraft.internal.states.PullState(
            self.property_names, self.part_properties, self.project
        )

        self.assertTrue(self.state == other, "Expected states to be identical")

    def test_properties_of_interest(self):
        self.part_properties.update(
            {
                "override-pull": "touch override-pull",
                "plugin": "test-plugin",
                "parse-info": "test-parse-info",
                "stage-packages": ["test-stage-package"],
                "source": "test-source",
                "source-commit": "test-source-commit",
                "source-depth": "test-source-depth",
                "source-tag": "test-source-tag",
                "source-type": "test-source-type",
                "source-branch": "test-source-branch",
                "source-subdir": "test-source-subdir",
            }
        )

        properties = self.state.properties_of_interest(self.part_properties)
        self.assertThat(len(properties), Equals(12))
        self.assertThat(properties["foo"], Equals("bar"))
        self.assertThat(properties["override-pull"], Equals("touch override-pull"))
        self.assertThat(properties["plugin"], Equals("test-plugin"))
        self.assertThat(properties["parse-info"], Equals("test-parse-info"))
        self.assertThat(properties["stage-packages"], Equals(["test-stage-package"]))
        self.assertThat(properties["source"], Equals("test-source"))
        self.assertThat(properties["source-commit"], Equals("test-source-commit"))
        self.assertThat(properties["source-depth"], Equals("test-source-depth"))
        self.assertThat(properties["source-tag"], Equals("test-source-tag"))
        self.assertThat(properties["source-type"], Equals("test-source-type"))
        self.assertThat(properties["source-branch"], Equals("test-source-branch"))
        self.assertThat(properties["source-subdir"], Equals("test-source-subdir"))

    def test_project_options_of_interest(self):
        options = self.state.project_options_of_interest(self.project)

        self.assertThat(len(options), Equals(1))
        self.assertThat(options["deb_arch"], Equals("amd64"))


def test_comparison_not_equal(pull_state, pull_state_variant):
    assert pull_state != pull_state_variant



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

from snapcraft.internal.errors import SnapcraftException


class SnapcraftDatastoreVersionUnsupported(SnapcraftException):
    def __init__(
        self, *, path: pathlib.Path, current_version: int, supported_version: int
    ) -> None:
        self.path = path
        self.current_version = current_version
        self.supported_version = supported_version

    def get_brief(self) -> str:
        return f"This version of snapcraft does not support version {self.current_version!r} of the {self.path} datastore."

    def get_resolution(self) -> str:
        return "Use 'snap revert' or 'snap refresh' to install previously used version of snapcraft (or newer)."



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from snapcraft.file_utils import calculate_hash

from . import errors


def split_checksum(source_checksum: str) -> Tuple:
    """Splits source_checksum into algorithm and hash.
    :raises ValueError: if source_checksum is not of the form algorithm/hash.
    :returns: a tuple consisting of the algorithm and the hash.
    """
    try:
        algorithm, digest = source_checksum.split("/", 1)

    except ValueError:
        raise ValueError("invalid checksum format: {!r}".format(source_checksum))
    return (algorithm, digest)


def verify_checksum(source_checksum: str, checkfile: str) -> Tuple:
    """Verifies that checkfile corresponds to the given source_checksum.
    :param str source_checksum: algorithm/hash expected for checkfile.
    :param str checkfile: the file to calculate the sum for with the
                          algorithm defined in source_checksum.
    :raises ValueError: if source_checksum is not of the form algorithm/hash.
    :raises DigestDoesNotMatchError: if checkfile does not match the expected
                                     hash calculated with the algorithm defined
                                     in source_checksum.
    :returns: a tuple consisting of the algorithm and the hash.
    """
    algorithm, digest = split_checksum(source_checksum)

    calculated_digest = calculate_hash(checkfile, algorithm=algorithm)
    if digest != calculated_digest:
        raise errors.DigestDoesNotMatchError(digest, calculated_digest)

    return (algorithm, digest)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys
from unittest import mock

from testtools.matchers import Equals, FileExists, MatchesRegex

from snapcraft.internal import sources
from tests import unit
from tests.file_utils import get_snapcraft_path


class TestRpm(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.rpm_file_path = os.path.join(
            get_snapcraft_path(),
            "tests",
            "integration",
            "snaps",
            "rpm-hello",
            "small-0.1-1.noarch.rpm",
        )

        self.dest_dir = "dst"
        os.makedirs(self.dest_dir)

    def test_pull_rpm_file_must_extract(self):
        rpm_source = sources.Rpm(self.rpm_file_path, self.dest_dir)
        rpm_source.pull()

        self.assertThat(os.path.join(self.dest_dir, "bin", "hello"), FileExists())

    @mock.patch.object(sources.Rpm, "provision")
    def test_pull_rpm_must_not_clean_targets(self, mock_provision):
        rpm_source = sources.Rpm(self.rpm_file_path, self.dest_dir)
        rpm_source.pull()

        mock_provision.assert_called_once_with(
            self.dest_dir,
            clean_target=False,
            src=os.path.join(self.dest_dir, "small-0.1-1.noarch.rpm"),
        )

    def test_has_source_handler_entry_on_linux(self):
        if sys.platform == "linux":
            self.assertTrue(sources._source_handler["rpm"] is sources.Rpm)
        else:
            self.assertRaises(KeyError, sources._source_handler["rpm"])

    @mock.patch("subprocess.check_output")
    def test_pull_failure(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, [])

        rpm_source = sources.Rpm(self.rpm_file_path, self.dest_dir)
        raised = self.assertRaises(sources.errors.SnapcraftPullError, rpm_source.pull)
        self.assertThat(
            raised.command,
            MatchesRegex("rpm2cpio .*/dst/small-0.1-1.noarch.rpm | cpio -idmv"),
        )
        self.assertThat(raised.exit_code, Equals(1))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Tim Süberkrüb
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import FileExists

from tests import integration


class SevenZipTestCase(integration.TestCase):

    _7z_test_files = {"test1.txt", "test2.txt", "test3.txt"}

    def test_stage_7z(self):
        self.run_snapcraft("stage", "7z-hello")

        for filename in self._7z_test_files:
            self.assertThat(os.path.join(self.stage_dir, filename), FileExists())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from typing import Dict, Any
from unittest import mock

import pytest

from snapcraft.storeapi.v2 import validation_sets
from snapcraft.storeapi import StoreClient


@pytest.fixture
def validation_sets_payload():
    return validation_sets.ValidationSets.unmarshal(
        {
            "assertions": [
                {
                    "headers": {
                        "account-id": "AccountIDXXXOfTheRequestingUserX",
                        "authority-id": "AccountIDXXXOfTheRequestingUserX",
                        "name": "certification-x1",
                        "revision": "222",
                        "sequence": "9",
                        "series": "16",
                        "sign-key-sha3-384": "XSignXKeyXHashXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                        "snaps": [
                            {
                                "id": "XXSnapIDForXSnapName1XXXXXXXXXXX",
                                "name": "snap-name-1",
                                "presence": "optional",
                            },
                            {
                                "id": "XXSnapIDForXSnapName2XXXXXXXXXXX",
                                "name": "snap-name-2",
                            },
                        ],
                        "timestamp": "2020-10-29T16:36:56Z",
                        "type": "validation-set",
                    }
                },
            ]
        }
    )


@pytest.fixture(autouse=True)
def fake_dashboard_post_validation_sets(validation_sets_payload):
    patched_get_validation_sets = mock.patch.object(
        StoreClient, "post_validation_sets", return_value=validation_sets_payload
    )
    yield patched_get_validation_sets.start()
    patched_get_validation_sets.stop()


@pytest.fixture(autouse=True)
def fake_dashboard_post_validation_sets_build_assertion(validation_sets_payload):
    patched_get_validation_sets = mock.patch.object(
        StoreClient,
        "post_validation_sets_build_assertion",
        return_value=validation_sets_payload.assertions[0],
    )
    yield patched_get_validation_sets.start()
    patched_get_validation_sets.stop()


@pytest.fixture(autouse=True)
def fake_dashboard_get_validation_sets(validation_sets_payload):
    patched_get_validation_sets = mock.patch.object(
        StoreClient, "get_validation_sets", return_value=validation_sets_payload
    )
    yield patched_get_validation_sets.start()
    patched_get_validation_sets.stop()


@pytest.fixture(autouse=True)
def fake_snap_sign():
    def sign(assertion: Dict[str, Any], *, key_name: str) -> bytes:
        return (json.dumps(assertion) + f"\n\nSIGNED{key_name}").encode()

    patched_snap_sign = mock.patch(
        "snapcraft.cli.assertions._sign_assertion", side_effect=sign
    )
    yield patched_snap_sign.start()
    patched_snap_sign.stop()


@pytest.mark.usefixtures("mock_subprocess_run")
def test_edit_validation_sets_with_no_changes_to_existing_set(
    click_run,
    fake_dashboard_get_validation_sets,
    fake_dashboard_post_validation_sets_build_assertion,
    fake_dashboard_post_validation_sets,
    fake_snap_sign,
):
    cmd = [
        "edit-validation-sets",
        "AccountIDXXXOfTheRequestingUserX",
        "certification-x1",
        "9",
    ]

    result = click_run(cmd)

    assert result.exit_code == 0
    assert result.output.strip() == "No changes made."
    fake_dashboard_get_validation_sets.assert_called_once_with(
        name="certification-x1", sequence="9"
    )
    fake_dashboard_post_validation_sets_build_assertion.assert_not_called()
    fake_snap_sign.assert_not_called()
    fake_dashboard_post_validation_sets.assert_not_called()


@pytest.fixture
def fake_edit_validation_sets():
    patched_edit_validation_sets = mock.patch(
        "snapcraft.cli.assertions._edit_validation_sets"
    )
    yield patched_edit_validation_sets.start()
    patched_edit_validation_sets.stop()


@pytest.mark.parametrize("key_name", [None, "general", "main"])
def test_edit_validation_sets_with_changes_to_existing_set(
    click_run,
    validation_sets_payload,
    fake_dashboard_get_validation_sets,
    fake_dashboard_post_validation_sets_build_assertion,
    fake_dashboard_post_validation_sets,
    fake_snap_sign,
    fake_edit_validation_sets,
    key_name,
):
    cmd = [
        "edit-validation-sets",
        "AccountIDXXXOfTheRequestingUserX",
        "certification-x1",
        "9",
    ]
    if key_name is not None:
        cmd.extend(["--key-name", key_name])

    fake_edit_validation_sets.return_value = {
        "account-id": "AccountIDXXXOfTheRequestingUserX",
        "name": "certification-x1",
        "sequence": "9",
        "snaps": [
            {
                "name": "snap-name-1",
                "id": "XXSnapIDForXSnapName1XXXXXXXXXXX",
                "presence": "required",
            },
            {"name": "snap-name-2", "id": "XXSnapIDForXSnapName2XXXXXXXXXXX"},
        ],
    }

    result = click_run(cmd)

    fake_dashboard_get_validation_sets.assert_called_once_with(
        name="certification-x1", sequence="9"
    )
    fake_dashboard_post_validation_sets_build_assertion.assert_called_once_with(
        validation_sets={
            "account-id": "AccountIDXXXOfTheRequestingUserX",
            "name": "certification-x1",
            "sequence": "9",
            "snaps": [
                {
                    "name": "snap-name-1",
                    "id": "XXSnapIDForXSnapName1XXXXXXXXXXX",
                    "presence": "required",
                },
                {"name": "snap-name-2", "id": "XXSnapIDForXSnapName2XXXXXXXXXXX"},
            ],
        }
    )
    fake_dashboard_post_validation_sets.assert_called_once_with(
        signed_validation_sets=(
            '{"account-id": "AccountIDXXXOfTheRequestingUserX", "authority-id": '
            '"AccountIDXXXOfTheRequestingUserX", "name": "certification-x1", '
            '"revision": "222", "sequence": "9", "snaps": [{"name": "snap-name-1", '
            '"id": "XXSnapIDForXSnapName1XXXXXXXXXXX", "presence": "optional"}, '
            '{"name": "snap-name-2", "id": "XXSnapIDForXSnapName2XXXXXXXXXXX"}], '
            '"timestamp": "2020-10-29T16:36:56Z", "type": "validation-set", '
            f'"series": "16"}}\n\nSIGNED{key_name}'
        ).encode()
    )
    fake_snap_sign.assert_called_once_with(
        validation_sets_payload.assertions[0].marshal(), key_name=key_name,
    )
    assert result.exit_code == 0
    assert result.output.strip() == ""



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import textwrap

import fixtures

import snapcraft
from snapcraft.internal import project_loader
from tests import unit


class LifecycleTestBase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)
        self.project_options = snapcraft.ProjectOptions()

        self.fake_install_build_packages = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_packages",
            return_value=list(),
        )
        self.useFixture(self.fake_install_build_packages)

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.project_loader._config.Config.get_build_packages",
                return_value=set(),
            )
        )

        self.fake_install_build_snaps = fixtures.MockPatch(
            "snapcraft.internal.lifecycle._runner._install_build_snaps",
            return_value=list(),
        )
        self.useFixture(self.fake_install_build_snaps)

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.project_loader._config.Config.get_build_snaps",
                return_value=set(),
            )
        )

    def make_snapcraft_project(self, parts, snap_type=""):
        yaml = textwrap.dedent(
            """\
            name: test
            base: core18
            version: "1.0"
            summary: test
            description: test
            confinement: strict
            grade: stable
            {type}

            {parts}
            """
        )

        self.snapcraft_yaml_file_path = self.make_snapcraft_yaml(
            yaml.format(parts=parts, type=snap_type)
        )
        project = snapcraft.project.Project(
            snapcraft_yaml_file_path=self.snapcraft_yaml_file_path
        )
        return project_loader.load_config(project)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import DirExists, FileExists, Not

from tests import integration


class LibraryPrecedenceTestCase(integration.TestCase):
    def test_snapped_library_takes_precedence_over_system(self):
        self.run_snapcraft("stage", "fake-curl-library")
        self.run_snapcraft(["prime", "main", "fake-curl"])

        # We will have everything in, given that we require dependencies
        # to be primed.
        self.assertThat(os.path.join(self.prime_dir, "bin", "main"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "lib"), DirExists())
        # If this exist, snapcraft brought libbcurl in from the host.
        self.assertThat(os.path.join(self.prime_dir, "usr"), Not(DirExists()))

        # Prime the rest of the way.
        self.run_snapcraft("prime")

        # Now verify the lib we got was the one from the snap, not from the
        # system.
        self.assertThat(os.path.join(self.prime_dir, "lib", "libcurl.so"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "usr"), Not(DirExists()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import DirExists, FileContains, Not

from tests import integration


class FilesetsTestCase(integration.TestCase):
    def test_filesets(self):
        self.run_snapcraft("snap", "make-with-filesets")

        expected_dirs = (
            os.path.join(self.stage_dir, "share"),
            os.path.join(self.stage_dir, "bin"),
            os.path.join(self.prime_dir, "bin"),
        )
        for expected_dir in expected_dirs:
            self.assertThat(expected_dir, DirExists())

        self.assertThat(os.path.join(self.prime_dir, "share"), Not(DirExists()))

        expected_files = (
            (os.path.join(self.stage_dir, "share", "share1"), "share1\n"),
            (os.path.join(self.stage_dir, "share", "share2"), "share2\n"),
            (os.path.join(self.stage_dir, "bin", "bin1"), "bin1\n"),
            (os.path.join(self.stage_dir, "bin", "bin2"), "bin2\n"),
            (os.path.join(self.prime_dir, "bin", "bin1"), "bin1\n"),
            (os.path.join(self.prime_dir, "bin", "bin2"), "bin2\n"),
        )
        for path, content in expected_files:
            self.assertThat(path, FileContains(content))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import requests
from testtools.matchers import Contains, Equals

from snapcraft.internal.sources import _base, errors
from tests import unit


class TestFileBase(unit.TestCase):
    def get_mock_file_base(self, source, dir):
        file_src = _base.FileBase(source, dir)
        setattr(file_src, "provision", mock.Mock())
        return file_src

    @mock.patch("snapcraft.internal.sources._base.FileBase.download")
    def test_pull_url(self, mock_download):
        mock_download.return_value = "dir"
        file_src = self.get_mock_file_base("http://snapcraft.io/snapcraft.yaml", "dir")
        file_src.pull()

        mock_download.assert_called_once_with()
        file_src.provision.assert_called_once_with(
            file_src.source_dir, src="dir", clean_target=False
        )

    @mock.patch("shutil.copy2")
    def test_pull_copy(self, mock_shutil_copy2):
        file_src = self.get_mock_file_base("snapcraft.yaml", "dir")
        file_src.pull()

        expected = os.path.join(file_src.source_dir, "snapcraft.yaml")
        mock_shutil_copy2.assert_called_once_with(file_src.source, expected)
        file_src.provision.assert_called_once_with(
            file_src.source_dir, src=expected, clean_target=False
        )

    @mock.patch("shutil.copy2", side_effect=FileNotFoundError())
    def test_pull_copy_source_does_not_exist(self, mock_shutil_copy2):
        file_src = self.get_mock_file_base("does-not-exist.tar.gz", ".")

        raised = self.assertRaises(errors.SnapcraftSourceNotFoundError, file_src.pull)

        self.assertThat(
            str(raised), Contains("Failed to pull source: 'does-not-exist.tar.gz'")
        )

    @mock.patch("snapcraft.internal.sources._base.requests")
    @mock.patch("snapcraft.internal.sources._base.download_requests_stream")
    @mock.patch("snapcraft.internal.sources._base.download_urllib_source")
    def test_download_file_destination(self, dus, drs, req):
        file_src = self.get_mock_file_base("http://snapcraft.io/snapcraft.yaml", "dir")
        self.assertFalse(hasattr(file_src, "file"))

        file_src.pull()

        self.assertThat(
            file_src.file,
            Equals(
                os.path.join(file_src.source_dir, os.path.basename(file_src.source))
            ),
        )

    @mock.patch("snapcraft.internal.common.get_url_scheme", return_value=False)
    @mock.patch("requests.get", side_effect=requests.exceptions.ConnectionError("foo"))
    def test_download_error(self, mock_get, mock_gus):
        base = self.get_mock_file_base("", "")
        base.source_checksum = False

        raised = self.assertRaises(errors.SnapcraftRequestError, base.download, None)

        self.assertThat(str(raised), Contains("Network request error"))

    @mock.patch("snapcraft.internal.sources._base.download_requests_stream")
    @mock.patch("snapcraft.internal.sources._base.requests")
    def test_download_http(self, mock_requests, mock_download):
        file_src = self.get_mock_file_base("http://snapcraft.io/snapcraft.yaml", "dir")

        mock_request = mock.Mock()
        mock_requests.get.return_value = mock_request

        file_src.pull()

        mock_requests.get.assert_called_once_with(
            file_src.source, stream=True, allow_redirects=True
        )
        mock_request.raise_for_status.assert_called_once_with()
        mock_download.assert_called_once_with(mock_request, file_src.file)

    @mock.patch("snapcraft.internal.sources._base.download_urllib_source")
    def test_download_ftp(self, mock_download):
        file_src = self.get_mock_file_base("ftp://snapcraft.io/snapcraft.yaml", "dir")

        file_src.pull()

        mock_download.assert_called_once_with(file_src.source, file_src.file)

    @mock.patch("snapcraft.internal.indicators.urlretrieve")
    def test_download_ftp_url_opener(self, mock_urlretrieve):
        file_src = self.get_mock_file_base("ftp://snapcraft.io/snapcraft.yaml", "dir")

        file_src.pull()

        self.assertThat(mock_urlretrieve.call_count, Equals(1))
        self.assertThat(mock_urlretrieve.call_args[0][0], Equals(file_src.source))
        self.assertThat(mock_urlretrieve.call_args[0][1], Equals(file_src.file))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from time import sleep
from typing import Any, Dict, Iterable, List, Optional, TextIO, Union

import requests

from snapcraft.internal.indicators import download_requests_stream
from . import _upload, errors, http_clients
from ._dashboard_api import DashboardAPI
from ._snap_api import SnapAPI
from ._up_down_client import UpDownClient
from .constants import DEFAULT_SERIES
from .v2 import channel_map, releases, validation_sets, whoami


logger = logging.getLogger(__name__)


class StoreClient:
    """High-level client Snap resources."""

    @property
    def use_candid(self) -> bool:
        return isinstance(self.auth_client, http_clients.CandidClient)

    def __init__(self, use_candid: bool = False) -> None:
        super().__init__()

        self.client = http_clients.Client()

        candid_has_credentials = http_clients.CandidClient.has_credentials()
        logger.debug(
            f"Candid forced: {use_candid}. Candid crendendials: {candid_has_credentials}."
        )
        if use_candid or candid_has_credentials:
            self.auth_client: http_clients.AuthClient = http_clients.CandidClient()
        else:
            self.auth_client = http_clients.UbuntuOneAuthClient()

        self.snap = SnapAPI(self.client)
        self.dashboard = DashboardAPI(self.auth_client)
        self._updown = UpDownClient(self.client)

    def login(
        self,
        *,
        acls: Iterable[str] = None,
        channels: Iterable[str] = None,
        packages: Iterable[Dict[str, str]] = None,
        expires: str = None,
        config_fd: TextIO = None,
        **kwargs,
    ) -> None:
        if config_fd is not None:
            return self.auth_client.login(config_fd=config_fd, **kwargs)

        if acls is None:
            acls = [
                "package_access",
                "package_manage",
                "package_push",
                "package_register",
                "package_release",
                "package_update",
            ]

        macaroon = self.dashboard.get_macaroon(
            acls=acls, packages=packages, channels=channels, expires=expires,
        )
        self.auth_client.login(macaroon=macaroon, **kwargs)

    def export_login(self, *, config_fd: TextIO, encode=False) -> None:
        self.auth_client.export_login(config_fd=config_fd, encode=encode)

    def logout(self):
        self.auth_client.logout()

    def whoami(self) -> whoami.WhoAmI:
        """Return user relevant login information."""
        return self.dashboard.whoami()

    def acl(self) -> Dict[str, Any]:
        """Return permissions for the logged-in user."""

        acl_data = {}

        acl_info = self.dashboard.verify_acl()
        for key in ("snap_ids", "channels", "permissions", "expires"):
            acl_data[key] = acl_info.get(key)

        return acl_data

    def get_snap_name_for_id(self, snap_id: str) -> str:
        declaration_assertion = self.snap.get_assertion("snap-declaration", snap_id)
        return declaration_assertion["headers"]["snap-name"]

    def verify_acl(self) -> Dict[str, Union[List[str], str]]:
        return self.dashboard.verify_acl()

    def get_account_information(self):
        return self.dashboard.get_account_information()

    def register_key(self, account_key_request):
        return self.dashboard.register_key(account_key_request)

    def register(self, snap_name: str, is_private: bool = False, store_id: str = None):
        return self.dashboard.register(
            snap_name, is_private=is_private, store_id=store_id, series=DEFAULT_SERIES,
        )

    def upload_precheck(self, snap_name):
        return self.dashboard.snap_upload_precheck(snap_name)

    def push_snap_build(self, snap_id, snap_build):
        return self.dashboard.push_snap_build(snap_id, snap_build)

    def upload(
        self,
        snap_name,
        snap_filename,
        delta_format=None,
        source_hash=None,
        target_hash=None,
        delta_hash=None,
        built_at=None,
        channels: Optional[List[str]] = None,
    ):
        updown_data = _upload.upload_files(snap_filename, self._updown)

        return self.dashboard.snap_upload_metadata(
            snap_name,
            updown_data,
            delta_format=delta_format,
            source_hash=source_hash,
            target_hash=target_hash,
            delta_hash=delta_hash,
            built_at=built_at,
            channels=channels,
        )

    def release(
        self,
        snap_name,
        revision,
        channels,
        progressive_percentage: Optional[int] = None,
    ):
        return self.dashboard.snap_release(
            snap_name,
            revision,
            channels,
            progressive_percentage=progressive_percentage,
        )

    def get_snap_status(self, snap_name, arch=None):
        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name=snap_name, arch=arch)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        response = self.dashboard.snap_status(snap_id, DEFAULT_SERIES, arch)

        if not response:
            raise errors.SnapNotFoundError(snap_name=snap_name, arch=arch)

        return response

    def get_snap_channel_map(self, *, snap_name: str) -> channel_map.ChannelMap:
        return self.dashboard.get_snap_channel_map(snap_name=snap_name)

    def get_snap_releases(self, *, snap_name: str) -> releases.Releases:
        return self.dashboard.get_snap_releases(snap_name=snap_name)

    def post_validation_sets_build_assertion(
        self, *, validation_sets: Dict[str, Any]
    ) -> validation_sets.BuildAssertion:
        return self.dashboard.post_validation_sets_build_assertion(validation_sets)

    def post_validation_sets(
        self, *, signed_validation_sets: bytes
    ) -> validation_sets.ValidationSets:
        return self.dashboard.post_validation_sets(signed_validation_sets)

    def get_validation_sets(
        self, *, name: Optional[str] = None, sequence: Optional[str] = None
    ) -> validation_sets.ValidationSets:
        return self.dashboard.get_validation_sets(name=name, sequence=sequence)

    def close_channels(self, snap_id, channel_names):
        return self.dashboard.close_channels(snap_id, channel_names)

    def download(
        self,
        snap_name,
        *,
        risk: str,
        download_path: str,
        track: Optional[str] = None,
        arch: Optional[str] = None,
        except_hash: str = "",
    ):
        snap_info = self.snap.get_info(snap_name)
        channel_mapping = snap_info.get_channel_mapping(
            risk=risk, track=track, arch=arch
        )
        if channel_mapping.download.sha3_384 == except_hash:
            return channel_mapping.download.sha3_384

        try:
            channel_mapping.download.verify(download_path)
        except errors.StoreDownloadError:
            self._download_snap(channel_mapping.download, download_path)

        channel_mapping.download.verify(download_path)
        return channel_mapping.download.sha3_384

    def _download_snap(self, download_details, download_path):
        # we only resume when redirected to our CDN since we use internap's
        # special sauce.
        total_read = 0
        probe_url = requests.head(download_details.url)
        if probe_url.is_redirect and "internap" in probe_url.headers["Location"]:
            download_url = probe_url.headers["Location"]
            resume_possible = True
        else:
            download_url = download_details.url
            resume_possible = False

        # HttpAdapter cannot help here as this is a stream.
        # LP: #1617765
        not_downloaded = True
        retry_count = 5
        while not_downloaded and retry_count:
            headers = {}
            if resume_possible and os.path.exists(download_path):
                total_read = os.path.getsize(download_path)
                headers["Range"] = "bytes={}-".format(total_read)
            request = self.client.request(
                "GET", download_url, headers=headers, stream=True
            )
            request.raise_for_status()
            redirections = [h.headers["Location"] for h in request.history]
            if redirections:
                logger.debug(
                    "Redirections for {!r}: {}".format(
                        download_url, ", ".join(redirections)
                    )
                )
            try:
                download_requests_stream(request, download_path, total_read=total_read)
                not_downloaded = False
            except requests.exceptions.ChunkedEncodingError as e:
                logger.debug(
                    "Error while downloading: {!r}. "
                    "Retries left to download: {!r}.".format(e, retry_count)
                )
                retry_count -= 1
                if not retry_count:
                    raise e
                sleep(1)

    def push_assertion(self, snap_id, assertion, endpoint, force=False):
        return self.dashboard.push_assertion(snap_id, assertion, endpoint, force)

    def get_assertion(self, snap_id, endpoint, params=None):
        return self.dashboard.get_assertion(snap_id, endpoint, params=params)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        return self.dashboard.sign_developer_agreement(latest_tos_accepted)

    def upload_metadata(self, snap_name, metadata, force):
        """Upload the metadata to the server."""
        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name=snap_name)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self.dashboard.upload_metadata(snap_id, snap_name, metadata, force)

    def upload_binary_metadata(self, snap_name, metadata, force):
        """Upload the binary metadata to the server."""
        account_info = self.get_account_information()
        try:
            snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
        except KeyError:
            raise errors.SnapNotFoundError(snap_name=snap_name)

        if snap_id is None:
            raise errors.NoSnapIdError(snap_name)

        return self.dashboard.upload_binary_metadata(
            snap_id, snap_name, metadata, force
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import urllib.parse

from pyramid import response

import tests
from tests.fake_servers import base

logger = logging.getLogger(__name__)


class FakeStoreSearchServer(base.BaseFakeServer):

    # XXX This fake server as reused as download server, to avoid passing a
    # port as an argument. --elopio - 2016-05-01

    def configure(self, configurator):
        configurator.add_route("info", "/v2/snaps/info/{snap}", request_method="GET")
        configurator.add_view(self.info, route_name="info")

        configurator.add_route(
            "download", "/download-snap/{snap}", request_method="GET"
        )
        configurator.add_view(self.download, route_name="download")

    def info(self, request):
        snap = request.matchdict["snap"]
        logger.debug(
            "Handling details request for package {}, with headers {}".format(
                snap, request.headers
            )
        )
        if "User-Agent" not in request.headers:
            response_code = 500
            return response.Response(None, response_code)
        payload = self._get_info_payload(request)
        if payload is None:
            response_code = 404
            return response.Response(json.dumps({}).encode(), response_code)
        response_code = 200
        content_type = "application/hal+json"
        return response.Response(
            payload, response_code, [("Content-Type", content_type)]
        )

    def _get_info_payload(self, request):
        # core snap is used in integration tests with fake servers.
        snap = request.matchdict["snap"]
        # tests/data/test-snap.snap
        test_sha3_384 = (
            "8c0118831680a22090503ee5db98c88dd90ef551d80fc816"
            "dec968f60527216199dacc040cddfe5cec6870db836cb908"
        )
        revision = "10000"
        confinement = "strict"

        if snap in ("snap-1", "test-snap", "core"):
            sha3_384 = test_sha3_384
        elif snap == "snapcraft":
            sha3_384 = test_sha3_384
            revision = "25"
            confinement = "classic"
        elif snap == "test-snap-with-wrong-sha":
            sha3_384 = "wrong sha"
        elif (
            snap == "test-snap-branded-store"
            and request.headers.get("Snap-Device-Store") == "Test-Branded"
        ):
            sha3_384 = test_sha3_384
        else:
            return None

        channel_map = list()
        for arch in ("amd64", "i386", "s390x", "arm64", "armhf", "ppc64el"):
            for risk in ("stable", "edge"):
                channel_map.append(
                    {
                        "channel": {
                            "architecture": arch,
                            "name": risk,
                            "released-at": "019-01-17T15:01:26.537392+00:00",
                            "risk": risk,
                            "track": "latest",
                        },
                        "download": {
                            "deltas": [],
                            "sha3-384": sha3_384,
                            "url": urllib.parse.urljoin(
                                "http://localhost:{}".format(self.server.server_port),
                                "download-snap/test-snap.snap",
                            ),
                        },
                        "created-at": "2019-01-16T14:59:16.711111+00:00",
                        "confinement": confinement,
                        "revision": revision,
                    }
                )

        if snap == "snap-1":
            snap_id = "snap-id-1"
        else:
            snap_id = f"{snap}-snap-id"

        return json.dumps(
            {
                "channel-map": channel_map,
                "snap": {
                    "name": snap,
                    "snap-id": snap_id,
                    "publisher": {
                        "id": snap + "-developer-id",
                        "validation": "unproven",
                    },
                },
                "snap-id": snap_id,
                "name": snap,
            }
        ).encode()

    def download(self, request):
        snap = request.matchdict["snap"]
        logger.debug("Handling download request for snap {}".format(snap))
        if "User-Agent" not in request.headers:
            response_code = 500
            return response.Response(None, response_code)
        response_code = 200
        content_type = "application/octet-stream"
        # TODO create a test snap during the test instead of hardcoding it.
        # --elopio - 2016-05-01
        snap_path = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap.snap"
        )

        with open(snap_path, "rb") as snap_file:
            return response.Response(
                snap_file.read(), response_code, [("Content-Type", content_type)]
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import textwrap

from testtools.matchers import FileContains, FileExists, Not

from snapcraft.internal import mangling
from tests import fixture_setup, unit


def _create_file(filename, contents):
    os.makedirs("test-dir", exist_ok=True)
    file_path = os.path.join("test-dir", filename)
    with open(file_path, "w") as f:
        f.write(contents)

    return file_path


class ManglingPythonShebangTestCase(unit.TestCase):
    def test_python(self):
        file_path = _create_file(
            "file",
            textwrap.dedent(
                """\
            #! /usr/bin/python2.7

            # Larger file
        """
            ),
        )
        mangling.rewrite_python_shebangs(os.path.dirname(file_path))
        self.assertThat(
            file_path,
            FileContains(
                textwrap.dedent(
                    """\
            #!/usr/bin/env python2.7

            # Larger file
        """
                )
            ),
        )

    def test_python3(self):
        file_path = _create_file("file", "#!/usr/bin/python3")
        mangling.rewrite_python_shebangs(os.path.dirname(file_path))
        self.assertThat(file_path, FileContains("#!/usr/bin/env python3"))

    def test_python_args(self):
        file_path1 = _create_file("file1", "#!/usr/bin/python")
        file_path2 = _create_file(
            "file2",
            textwrap.dedent(
                """\
            #! /usr/bin/python -E

            # Larger file
        """
            ),
        )
        mangling.rewrite_python_shebangs(os.path.dirname(file_path1))
        self.assertThat(file_path1, FileContains("#!/usr/bin/env python"))
        self.assertThat(
            file_path2,
            FileContains(
                textwrap.dedent(
                    """\
                #!/bin/sh
                ''''exec python -E -- "$0" "$@" # '''

                # Larger file
            """
                )
            ),
        )

    def test_python3_args(self):
        file_path1 = _create_file("file1", "#!/usr/bin/python3")
        file_path2 = _create_file("file2", "#!/usr/bin/python3 -E")
        mangling.rewrite_python_shebangs(os.path.dirname(file_path1))
        self.assertThat(file_path1, FileContains("#!/usr/bin/env python3"))
        self.assertThat(
            file_path2,
            FileContains(
                textwrap.dedent(
                    """\
                #!/bin/sh
                ''''exec python3 -E -- "$0" "$@" # '''"""
                )
            ),
        )

    def test_python_mixed_args(self):
        file_path1 = _create_file("file1", "#!/usr/bin/python")
        # Ensure extra spaces are chopped off
        file_path2 = _create_file("file2", "#!/usr/bin/python3       -Es")
        mangling.rewrite_python_shebangs(os.path.dirname(file_path1))
        self.assertThat(file_path1, FileContains("#!/usr/bin/env python"))
        self.assertThat(
            file_path2,
            FileContains(
                textwrap.dedent(
                    """\
                #!/bin/sh
                ''''exec python3 -Es -- "$0" "$@" # '''"""
                )
            ),
        )

    def test_following_docstring_no_rewrite(self):
        file_path = _create_file(
            "file",
            textwrap.dedent(
                """\
            #!/usr/bin/env python3.5

            '''
            This is a test
            =======================
        """
            ),
        )
        mangling.rewrite_python_shebangs(os.path.dirname(file_path))
        self.assertThat(
            file_path,
            FileContains(
                textwrap.dedent(
                    """\
            #!/usr/bin/env python3.5

            '''
            This is a test
            =======================
        """
                )
            ),
        )


class TestClearExecstack(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_elf = fixture_setup.FakeElf(root_path=self.path)
        self.useFixture(self.fake_elf)

    def test_execstack_clears(self):
        elf_files = [self.fake_elf["fake_elf-with-execstack"]]

        mangling.clear_execstack(elf_files=elf_files)

        self.assertThat("{}.execstack".format(elf_files[0].path), FileExists())

    def test_bad_execstack_does_not_blow_up(self):
        elf_files = [self.fake_elf["fake_elf-with-bad-execstack"]]

        mangling.clear_execstack(elf_files=elf_files)

        self.assertThat("{}.execstack".format(elf_files[0].path), Not(FileExists()))

    def test_no_execstack_does_nothing(self):
        elf_files = [self.fake_elf["fake_elf-2.23"]]

        mangling.clear_execstack(elf_files=elf_files)

        self.assertThat("{}.execstack".format(elf_files[0].path), Not(FileExists()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.errors import SnapcraftError, SnapcraftException

# dict of jsonschema validator -> cause pairs. Wish jsonschema just gave us
# better messages.


class MissingSnapcraftYamlError(SnapcraftError):

    fmt = (
        "Could not find {snapcraft_yaml_file_path}. Are you sure you are "
        "in the right directory?\n"
        "To start a new project, use `snapcraft init`"
    )

    def __init__(self, *, snapcraft_yaml_file_path):
        super().__init__(snapcraft_yaml_file_path=snapcraft_yaml_file_path)


class DuplicateSnapcraftYamlError(SnapcraftError):

    fmt = (
        "Found a {snapcraft_yaml_file_path!r} and a "
        "{other_snapcraft_yaml_file_path!r}.\n"
        "Please remove one and try again."
    )

    def __init__(
        self, *, snapcraft_yaml_file_path: str, other_snapcraft_yaml_file_path: str
    ) -> None:
        super().__init__(
            snapcraft_yaml_file_path=snapcraft_yaml_file_path,
            other_snapcraft_yaml_file_path=other_snapcraft_yaml_file_path,
        )


class SnapcraftExperimentalExtensionsRequiredError(SnapcraftException):
    def __init__(self, *, extension_name: str) -> None:
        self.extension_name = extension_name

    def get_brief(self) -> str:
        return f"Experimental extension {self.extension_name!r} is required, but not enabled."

    def get_resolution(self) -> str:
        return "This extension may be enabled with the '--enable-experimental-extensions' parameter."
